"""
NRV-Cellular Level simulations
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import sys

from ...backend.log_interface import pass_info, rise_error, rise_warning
from ...backend.MCore import *
from ...fmod.electrodes import *
from ...fmod.extracellular import *
from ...fmod.materials import *
from ...fmod.stimulus import *
from ...nmod.axons import *
from ...nmod.myelinated import *
from ...nmod.unmyelinated import *
from ..saving_handler import *
from .CL_discretization import *
from .CL_postprocessing import *

unmyelinated_models = [
    "HH",
    "Rattay_Aberham",
    "Sundt",
    "Tigerholm",
    "Schild_94",
    "Schild_97",
]
myelinated_models = ["MRG", "Gaines_motor", "Gaines_sensory"]


def firing_threshold_point_source(
    diameter,
    L,
    dist_elec,
    cath_time=100e-3,
    model="MRG",
    amp_max=2000,
    amp_tol=1,
    verbose=True,
    **kwargs,
):
    """
    Find the firing threshold with point source approximation using binary search.

    Parameters
    ----------
    diameter        : float
            axon diameter in um
    L               : float
            axon length in um
    dist_elec       : float
            y coordinate of the point source electrode in um
    cath_time      : float
            cathodic pulse width in ms
    model           : str
            Axon model
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    verbose         : bool
            set the verbosity of the search
    **kwargs:
            See below

    Keyword Arguments:
            material       :   str
                    material used to compute the extracellular field
            position_elect : float
                    x location of the electrode, between 0 and 1
            z_elect        : float
                    z coordinate of the electrode
            amp_min        : float
                    minimum amplitude for the binary search
            f_dlambda      : float
                    To set the number of segment with the d_dlambda rule (Hz)
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            anod_first  : bool
                    Set the anodic phase before the cathodic phase
            t_inter : float
                    Specify the interphase delay in ms
            cath_an_ratio: float
                    Specify the cathodic/anodic ratio
            dt              : float
                    Set the dt to a specific value in ms
            cath_an_ratio          : float
                    Set the tolorance for dt in %
            nseg            : int
                    Set the number of segment for the axon
            nseg_tol        : float
                    Set the nseg tolerance in %
            dt_tol        : float
                    Set the dt tolerance in %
            amp_tol_abs     : float
                    Set the absolute tolerance for the binary search in uA.
            node_shift      : float between -1 and 1
                    Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
            n_nodes         : integer
                    Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

    Returns
    -------
    threshold       : float
            estimated firing threshold in uA
    """

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        if model in myelinated_models:
            t_sim = 5
        else:
            t_sim = 20

    if "material" in kwargs:
        material = kwargs.get("material")
    else:
        material = material = "endoneurium_bhadra"

    if "anod_first" in kwargs:
        anod_first = kwargs.get("anod_first")
    else:
        anod_first = False

    if "t_inter" in kwargs:
        t_inter = kwargs.get("t_inter")
    else:
        t_inter = 0

    if "cath_an_ratio" in kwargs:
        cath_an_ratio = kwargs.get("cath_an_ratio")
    else:
        cath_an_ratio = 1

    if "position_elec" in kwargs:
        position_elec = kwargs.get("position_elec")
    else:
        position_elec = 0.5

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "f_dlambda" in kwargs:
        f_dlambda = kwargs.get("f_dlambda")
    else:
        f_dlambda = 0

    if "dt" in kwargs:
        dt = kwargs.get("dt")
    else:
        dt = 0

    if "dt_tol" in kwargs:
        dt_tol = kwargs.get("dt_tol")
    else:
        dt_tol = amp_tol

    if "nseg" in kwargs:
        nseg = kwargs.get("nseg")
    else:
        nseg = 0

    if "Nseg_per_sec" in kwargs:
        Nseg_per_sec = kwargs.get("Nseg_per_sec")
    else:
        Nseg_per_sec = 0

    if "nseg_tol" in kwargs:
        nseg_tol = kwargs.get("nseg_tol")
    else:
        nseg_tol = amp_tol

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "node_shift" in kwargs:
        node_shift = kwargs.get("node_shift")
    else:
        node_shift = 0

    if "n_nodes" in kwargs:
        n_nodes = kwargs.get("n_nodes")
    else:
        n_nodes = 0

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol
    # axon
    y = 0
    z = 0

    if dt == 0:
        if cath_an_ratio > 0 and cath_an_ratio < 1:
            dt = Get_dt(
                dt_tol,
                model,
                simulation_category="Spike_threshold",
                stim_pw=cath_time / cath_an_ratio,
            )
        else:
            dt = Get_dt(
                dt_tol, model, simulation_category="Spike_threshold", stim_pw=cath_time
            )

    if (nseg == 0) and (f_dlambda == 0) and (Nseg_per_sec == 0):
        if model in unmyelinated_models:
            Nseg_per_sec = Get_nseg(
                nseg_tol, model, simulation_category="Spike_threshold", d=diameter, L=L
            )
        else:
            Nseg_per_sec = Get_nseg(
                nseg_tol, model, simulation_category="Spike_threshold"
            )

    # extra cellular
    extra_material = load_material(material)
    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing firing current amplitude "
                + str(current_amp)
                + " uA"
            )
        # create axon
        if model in unmyelinated_models:
            if Nseg_per_sec:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, model=model, Nseg_per_sec=Nseg_per_sec
                )  # freq=f_dlambda
            else:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, freq=f_dlambda, model=model
                )
        elif model in myelinated_models:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            if Nseg_per_sec:
                axon1 = myelinated(
                    y,
                    z,
                    diameter,
                    L,
                    rec="nodes",
                    dt=dt,
                    model=model,
                    Nseg_per_sec=Nseg_per_sec,
                )  # freq=f_dlambda
            else:
                axon1 = myelinated(
                    y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
                )
        else:
            rise_error("Error: Specified model is not recognized.")
        # extra-cellular stimulation
        x_elec = L * position_elec
        if model in myelinated_models:  # Align electrode with node of Ranvier
            n_nodes = len(axon1.x_nodes)
            nearest_node = np.int32(np.round(position_elec * n_nodes))
            x_elec = axon1.x_nodes[nearest_node]
            if node_shift != 0:  # shift it if needed
                if node_shift > 0:
                    if nearest_node < n_nodes - 1:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec + d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
                else:
                    if nearest_node != 0:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec - d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )

        y_elec = dist_elec
        z_elec = 0
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)

        # stimulus def
        stim_1 = stimulus()
        start = 1
        I_cathod = current_amp
        if cath_an_ratio > 0:
            I_anod = I_cathod / cath_an_ratio

            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, I_anod, t_inter, anod_first=anod_first
            )
        else:
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, 0, 0, anod_first=anod_first
            )

        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=t_sim)
        del axon1
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        rasterize(results, "V_mem")
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if len(results["V_mem_raster_position"]) > 0:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not triggered")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    return current_amp


def firing_threshold_from_axon(
    axon, cath_time=100e-3, amp_max=2000, amp_tol=1, verbose=True, **kwargs
):
    """
    Find the firing threshold from a specified axon using binary search

    Parameters
    ----------
    axon        : axon class
            axon class
    cath_time      : float
            cathodic pulse width in ms
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    verbose         : bool
            set the verbosity of the search
    **kwargs:
            See below

    Keyword Arguments:
            amp_min        : float
                    minimum amplitude for the binary search
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            dt             : float
                    Set the dt to a specific value in ms
            anod_first  : bool
                    Set the anodic phase before the cathodic phase
            t_inter : float
                    Specify the interphase delay in ms
            cath_an_ratio: float
                    Specify the cathodic/anodic ratio
            cath_an_ratio          : float
                    Set the tolorance for dt in %
            amp_tol_abs     : float
                    Set the absolute tolerance for the binary search in uA.
            dt_tol        : float
                    Set the dt tolerance in %

    Returns
    -------
    threshold       : float
            estimated firing threshold in uA
    """

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 10

    if "anod_first" in kwargs:
        anod_first = kwargs.get("anod_first")
    else:
        anod_first = False

    if "t_inter" in kwargs:
        t_inter = kwargs.get("t_inter")
    else:
        t_inter = 0

    if "cath_an_ratio" in kwargs:
        cath_an_ratio = kwargs.get("cath_an_ratio")
    else:
        cath_an_ratio = 1

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "dt" in kwargs:
        dt = kwargs.get("dt")
    else:
        dt = 0

    if "dt_tol" in kwargs:
        dt_tol = kwargs.get("dt_tol")
    else:
        dt_tol = amp_tol

    if dt == 0:
        if cath_an_ratio > 0 and cath_an_ratio < 1:
            dt = Get_dt(
                dt_tol,
                axon.model,
                simulation_category="Spike_threshold",
                stim_pw=cath_time / cath_an_ratio,
            )
        else:
            dt = Get_dt(
                dt_tol,
                axon.model,
                simulation_category="Spike_threshold",
                stim_pw=cath_time,
            )

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1

    # axon.save_axon(save=True, fname='_axon.json', extracel_context=True)
    # del (axon)
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing firing current amplitude "
                + str(current_amp)
                + " uA"
            )

        # stimulus def
        stim_1 = stimulus()
        start = 1
        I_cathod = current_amp
        # I_cathod =
        if cath_an_ratio > 0:
            I_anod = I_cathod / cath_an_ratio
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, I_anod, t_inter, anod_first=anod_first
            )
        else:
            stim_1.biphasic_pulse(
                start, I_cathod, cath_time, 0, 0, anod_first=anod_first
            )
        # axon_load = load_any_axon(axon, extracel_context=True)
        # axon_th = copy.deepcopy(axon)
        axon.dt = dt
        axon.change_stimulus_from_elecrode(0, stim_1)

        # simulate axon activity
        results = axon.simulate(t_sim=t_sim, loaded_footprints=True)
        # del (axon)
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        rasterize(results, "V_mem")
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if len(results["V_mem_raster_position"]) > 0:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike triggered")
            amplitude_max_th = previous_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not triggered")
            current_amp = amplitude_max_th - delta_amp / 2
            amplitude_min_th = previous_amp

        if previous_amp == amp_max:
            current_amp = amp_min

        Niter += 1
    del axon
    return current_amp


def para_firing_threshold(
    diameter,
    L,
    material,
    dist_elec,
    cath_first=True,
    cath_time=60e-3,
    t_inter=40e-3,
    cath_an_ratio=5,
    position_elec=0.5,
    model="MRG",
    amp_max=2000,
    amp_min=0,
    amp_tol=1,
    verbose=False,
    f_dlambda=100,
    dt=0.005,
):
    if MCH.is_alone() or MCH.size < 3:
        if MCH.do_master_only_work():
            rise_error(
                "Error: parallel evaluation of a threshold by binary search needs at least 3 parallel processes, but only",
                MCH.size,
                "launched",
            )
        sys.exit(1)
    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol
    # axon
    y = 0
    z = 0
    # extra cellular
    extra_material = load_material(material)
    # Dichotomy initialization
    Niter = 1
    values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size)
    delta_amp = values[-1] - values[0]
    current_amp = (values[-1] - values[0]) / 2
    all_values = values
    while delta_amp > amplitude_tol or Niter == 1:
        # update toledance
        delta_amp = values[1] - values[0]
        # split job
        current_amp = values[MCH.rank]
        # create axon
        if model in unmyelinated_models:
            axon1 = unmyelinated(y, z, diameter, L, dt=dt, freq=f_dlambda, model=model)
        elif model in myelinated_models:
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        else:
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        # extra-cellular stimulation
        x_elec = L * position_elec
        y_elec = dist_elec
        z_elec = 0
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)

        # stimulus def
        stim_1 = stimulus()
        start = 1
        I_cathod = current_amp
        I_anod = I_cathod / cath_an_ratio
        stim_1.biphasic_pulse(
            start, I_cathod, cath_time, I_anod, t_inter, anod_first=(not cath_first)
        )

        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=5)
        del axon1
        # post-process results
        rasterize(results, "V_mem")
        # test simulation results, gather results to master
        if len(results["V_mem_raster_position"]) > 0:
            spike = np.asarray([True])
        else:
            spike = np.asarray([False])
        all_spikes = MCH.gather_jobs_as_array(spike)
        if MCH.do_master_only_work():
            # add the extrema (False at start, True at the end) already computed at the previous step if Niter > 1
            if Niter > 1:
                all_spikes = np.insert(np.append(all_spikes, True), 0, False)
            # find the zone where the spike appears, update min max
            appear_index = np.argmax(all_spikes == True)
            amplitude_max_th = all_values[appear_index]
            amplitude_min_th = all_values[appear_index - 1]
            # update current amp
            current_amp = amplitude_min_th + (amplitude_max_th - amplitude_min_th) / 2
            # compute new values
            all_values = np.linspace(
                amplitude_min_th, amplitude_max_th, num=MCH.size + 3
            )
            values = all_values[1:-1]
        # share new values
        values = MCH.master_broadcasts_array_to_all(values)
        Niter += 1
    # adapt Niter for correct return
    Niter -= 1
    # share master current amp to all process
    current_amp = MCH.master_broadcasts_array_to_all(current_amp)
    return current_amp, Niter


def blocking_threshold_point_source(
    diameter,
    L,
    dist_elec,
    block_freq,
    model="MRG",
    amp_max=2000,
    amp_tol=1,
    dt=0.005,
    Nseg_per_sec=2,
    verbose=True,
    **kwargs,
):
    """
    Find the blocking threshold with point source approximation using binary search.

    Parameters
    ----------
    diameter        : float
            axon diameter in um
    L               : float
            axon length in um
    dist_elec       : float
            y coordinate of the point source electrode in um
    block_freq      : float
            Blocking frequency in kHz
    model           : str
            Axon model
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    dt              : float
            time discretization in ms
    Nseg_per_sec    : int
            Number of segment per section (myelinated axons)
            Number of segment per mm of length (unmyelinated axons)
    verbose         : bool
            set the verbosity of the search

    **kwargs:
            See below

    Keyword Arguments:
            material       :   str
                    material used to compute the extracellular field
            position_elect : float
                    x location of the electrode, between 0 and 1
            z_elect        : float
                    z coordinate of the electrode
            amp_min        : float
                    minimum amplitude for the binary search
            f_dlambda      : float
                    To set the number of segment with the d_dlambda rule (Hz)
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            t_position     : float
                    Positition on the test electrode.
                    Relative position for unmyelinated axons.
                    Node number for myelinated axons.
            t_start        : float
                    start of the test pulse, in ms
            t_duration     : float
                    duration of the test pulse, in ms
            t_amplitude    : float
                    Amplitude of the test pulse, in nA
            b_start        : float
                    start of the blocking stimulation in ms
            b_duration     : float
                    duration of the blocking stimulation in ms
            node_shift      : float between -1 and 1
                    Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
            n_nodes         : integer
                    Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

    Returns
    -------
    threshold       : float
            estimated threshold in uA
    """

    if "material" in kwargs:
        material = kwargs.get("material")
    else:
        material = "endoneurium_bhadra"

    if "position_elec" in kwargs:
        position_elec = kwargs.get("position_elec")
    else:
        position_elec = 0.5

    if "z_elec" in kwargs:
        z_elec = kwargs.get("z_elec")
    else:
        z_elec = 0

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "f_dlambda" in kwargs:
        f_dlambda = kwargs.get("f_dlambda")
    else:
        f_dlambda = 100

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 40

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "t_position" in kwargs:
        t_position = kwargs.get("t_position")
    else:
        t_position = 0.05

    if "t_start" in kwargs:
        t_start = kwargs.get("t_start")
    else:
        t_start = 30

    if "t_duration" in kwargs:
        t_duration = kwargs.get("t_duration")
    else:
        t_duration = 1

    if "t_amplitude" in kwargs:
        t_amplitude = kwargs.get("t_amplitude")
    else:
        t_amplitude = 2

    if "b_start" in kwargs:
        b_start = kwargs.get("b_start")
    else:
        b_start = 3

    if "b_duration" in kwargs:
        b_duration = kwargs.get("b_duration")
    else:
        b_duration = t_sim

    if "node_shift" in kwargs:
        node_shift = kwargs.get("node_shift")
    else:
        node_shift = 0

    if "n_nodes" in kwargs:
        n_nodes = kwargs.get("n_nodes")
    else:
        n_nodes = 0

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    y = 0
    z = 0

    extra_material = load_material(material)
    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing block current amplitude "
                + str(current_amp)
                + " uA"
            )
        # create axon
        if model in unmyelinated_models:
            if Nseg_per_sec:
                axon1 = unmyelinated(
                    y,
                    z,
                    diameter,
                    L,
                    dt=dt,
                    Nseg_per_sec=Nseg_per_sec,
                    Nsec=1,
                    model=model,
                )
            else:
                axon1 = unmyelinated(
                    y, z, diameter, L, dt=dt, freq=f_dlambda, model=model
                )
        elif model in myelinated_models:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            if Nseg_per_sec:
                axon1 = myelinated(
                    y,
                    z,
                    diameter,
                    L,
                    rec="nodes",
                    dt=dt,
                    Nseg_per_sec=Nseg_per_sec,
                    model=model,
                )
            else:
                axon1 = myelinated(
                    y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
                )
        else:
            if n_nodes > 0:
                L = get_length_from_nodes(diameter, n_nodes)
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        # insert test spike
        axon1.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
        # extra-cellular stimulation
        x_elec = L * position_elec
        if model in myelinated_models:  # Align electrode with node of Ranvier
            n_nodes = len(axon1.x_nodes)
            nearest_node = np.int32(np.round(position_elec * n_nodes))
            x_elec = axon1.x_nodes[nearest_node]
            if node_shift != 0:  # shift it if needed
                if node_shift > 0:
                    if nearest_node < n_nodes - 1:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec + d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
                else:
                    if nearest_node != 0:
                        x_next_node = axon1.x_nodes[nearest_node + 1]
                        d_internode = np.abs(x_next_node - x_elec)
                        x_elec = x_elec - d_internode * node_shift
                    else:
                        rise_warning(
                            "Electrode is at the end of the axon and can not be further shifted."
                        )
        y_elec = dist_elec
        z_elec = z_elec
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)
        stim_1 = stimulus()
        stim_1.sinus(
            b_start, b_duration, current_amp, block_freq, dt=1 / (block_freq * 20)
        )
        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=t_sim)
        del axon1
        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        filter_freq(results, "V_mem", block_freq)
        rasterize(results, "V_mem_filtered", threshold=0)
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if block(results, t_start=t_start) == False:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not blocked")
            amplitude_min_th = current_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike blocked")
            amplitude_max_th = current_amp
            current_amp = amplitude_max_th - delta_amp / 2

        if previous_amp == amp_max:
            current_amp = amp_min
        Niter += 1
    return current_amp


def blocking_threshold_from_axon(
    axon, block_freq=10, amp_max=2000, amp_tol=1, dt=0.005, verbose=True, **kwargs
):
    """
    Find the blocking threshold with point source approximation using binary search.

    Parameters
    ----------
    diameter        : float
            axon diameter in um
    L               : float
            axon length in um
    dist_elec       : float
            y coordinate of the point source electrode in um
    block_freq      : float
            Blocking frequency in kHz
    model           : str
            Axon model
    amp_max         : float
            Maximum tested blocking amplitude for the binary search in uA
    amp_tol         : float
            Threshold tolerance in % for the binary search
    dt              : float
            time discretization in ms
    Nseg_per_sec    : int
            Number of segment per section (myelinated axons)
            Number of segment per mm of length (unmyelinated axons)
    verbose         : bool
            set the verbosity of the search

    **kwargs:
            See below

    Keyword Arguments:
            material       :   str
                    material used to compute the extracellular field
            position_elect : float
                    x location of the electrode, between 0 and 1
            z_elect        : float
                    z coordinate of the electrode
            amp_min        : float
                    minimum amplitude for the binary search
            f_dlambda      : float
                    To set the number of segment with the d_dlambda rule (Hz)
            t_sim          : float
                    Duration of the simulation in ms
            amp_tol_abs    : float
                    Specify an absolute amplitude tolerance for the binary search
                    in uA
            t_position     : float
                    Positition on the test electrode.
                    Relative position for unmyelinated axons.
                    Node number for myelinated axons.
            t_start        : float
                    start of the test pulse, in ms
            t_duration     : float
                    duration of the test pulse, in ms
            t_amplitude    : float
                    Amplitude of the test pulse, in nA
            b_start        : float
                    start of the blocking stimulation in ms
            b_duration     : float
                    duration of the blocking stimulation in ms
            node_shift      : float between -1 and 1
                    Align electrode with Node of Ranvier. When Shift = 0, Electrode is aligned with node
            n_nodes         : integer
                    Specify the number of Node of Ranvier for myelinated models. Overwrite L when specified.

    Returns
    -------
    threshold       : float
            estimated threshold in uA
    """

    if "amp_min" in kwargs:
        amp_min = kwargs.get("amp_min")
    else:
        amp_min = 0

    if "f_dlambda" in kwargs:
        f_dlambda = kwargs.get("f_dlambda")
    else:
        f_dlambda = 100

    if "t_sim" in kwargs:
        t_sim = kwargs.get("t_sim")
    else:
        t_sim = 40

    if "amp_tol_abs" in kwargs:
        amp_tol_abs = kwargs.get("amp_tol_abs")
    else:
        amp_tol_abs = 0

    if "t_position" in kwargs:
        t_position = kwargs.get("t_position")
    else:
        t_position = 0.05

    if "t_start" in kwargs:
        t_start = kwargs.get("t_start")
    else:
        t_start = 30

    if "t_duration" in kwargs:
        t_duration = kwargs.get("t_duration")
    else:
        t_duration = 1

    if "t_amplitude" in kwargs:
        t_amplitude = kwargs.get("t_amplitude")
    else:
        t_amplitude = 2

    if "b_start" in kwargs:
        b_start = kwargs.get("b_start")
    else:
        b_start = 3

    if "b_duration" in kwargs:
        b_duration = kwargs.get("b_duration")
    else:
        b_duration = t_sim

    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol

    # Dichotomy initialization
    previous_amp = amp_min
    delta_amp = np.abs(amp_max - amp_min)
    current_amp = amp_max
    Niter = 1
    keep_going = 1
    while keep_going:
        if verbose:
            pass_info(
                "Iteration number "
                + str(Niter)
                + ", testing block current amplitude "
                + str(current_amp)
                + " uA"
            )

        # insert test spike
        axon.insert_I_Clamp(t_position, t_start, t_duration, t_amplitude)
        # extra-cellular stimulation

        stim_1 = stimulus()
        stim_1.sinus(
            b_start, b_duration, current_amp, block_freq, dt=1 / (block_freq * 20)
        )

        axon.dt = dt
        axon.change_stimulus_from_elecrode(0, stim_1)

        # simulate axon activity
        results = axon.simulate(t_sim=t_sim, loaded_footprints=True)

        if verbose:
            pass_info(
                "... Iteration simulation performed in "
                + str(results["sim_time"])
                + " s"
            )
        # post-process results
        filter_freq(results, "V_mem", block_freq)
        rasterize(results, "V_mem_filtered", threshold=0)
        delta_amp = np.abs(current_amp - previous_amp)
        if amp_tol_abs > 0:
            if delta_amp < amp_tol_abs:
                keep_going = 0
            else:
                keep_going = 1
        else:
            if current_amp > previous_amp:
                tol = 100 * delta_amp / current_amp
            else:
                tol = 100 * delta_amp / previous_amp
            if tol <= amp_tol:
                keep_going = 0
            else:
                keep_going = 1
        previous_amp = current_amp
        # test simulation results, update dichotomy
        if block(results, t_start=t_start) == False:
            if current_amp == amp_max:
                rise_warning("Maximum Stimulation Current is too Low!")
                return 0
            if verbose:
                pass_info("... Spike not blocked")
            amplitude_min_th = current_amp
            current_amp = (delta_amp / 2) + amplitude_min_th
        else:
            if current_amp == amp_min:
                rise_warning("Minimum Stimulation Current is too High!")
                return 0
            if verbose:
                pass_info("... Spike blocked")
            amplitude_max_th = current_amp
            current_amp = amplitude_max_th - delta_amp / 2

        if previous_amp == amp_max:
            current_amp = amp_min
        Niter += 1
    return current_amp


def para_blocking_threshold(
    diameter,
    L,
    material,
    dist_elec,
    block_freq,
    position_elec=0.5,
    model="MRG",
    amp_max=2000,
    amp_min=0,
    amp_tol=15,
    verbose=True,
    f_dlambda=100,
    dt=0.005,
):
    amplitude_max_th = amp_max
    amplitude_min_th = amp_min
    amplitude_tol = amp_tol
    # axon
    y = 0
    z = 0
    # test spike
    t_start = 20
    duration = 0.1
    amplitude = 3
    # extra cellular
    extra_material = load_material(material)
    block_start = 3  # ms
    block_duration = 20  # ms
    # Dichotomy initialization
    Niter = 1
    values = np.linspace(amplitude_min_th, amplitude_max_th, num=MCH.size)
    delta_amp = values[-1] - values[0]
    current_amp = (values[-1] - values[0]) / 2
    all_values = values
    while delta_amp > amplitude_tol or Niter == 1:
        # update toledance
        delta_amp = values[1] - values[0]
        # split job
        current_amp = values[MCH.rank]
        # create axon
        if model in unmyelinated_models:
            axon1 = unmyelinated(y, z, diameter, L, dt=dt, freq=f_dlambda, model=model)
        elif model in myelinated_models:
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        else:
            axon1 = myelinated(
                y, z, diameter, L, rec="nodes", dt=dt, freq=f_dlambda, model=model
            )
        # insert test spike
        axon1.insert_I_Clamp(0, t_start, duration, amplitude)
        # extra-cellular stimulation
        x_elec = L * position_elec
        y_elec = dist_elec
        z_elec = 0
        elec_1 = point_source_electrode(x_elec, y_elec, z_elec)
        stim_1 = stimulus()
        stim_1.sinus(
            block_start,
            block_duration,
            current_amp,
            block_freq,
            dt=1 / (block_freq * 20),
        )
        stim_extra = stimulation(extra_material)
        stim_extra.add_electrode(elec_1, stim_1)
        axon1.attach_extracellular_stimulation(stim_extra)
        # simulate axon activity
        results = axon1.simulate(t_sim=25)
        del axon1
        # post-process results
        rasterize(results, "V_mem")

        # test simulation results, gather results to master
        blocked = [block(results)]
        all_blocks = MCH.gather_jobs_as_array(blocked)
        if MCH.do_master_only_work():
            # add the extrema (False at start, True at the end) already computed at the previous step if Niter > 1
            if Niter > 1:
                all_blocks = np.insert(np.append(all_blocks, True), 0, False)
            # find the zone where the blocking appears, update min max
            appear_index = np.argmax(all_blocks == True)
            amplitude_max_th = all_values[appear_index]
            amplitude_min_th = all_values[appear_index - 1]
            # update current amp
            current_amp = amplitude_min_th + (amplitude_max_th - amplitude_min_th) / 2
            # compute new values
            all_values = np.linspace(
                amplitude_min_th, amplitude_max_th, num=MCH.size + 3
            )
            values = all_values[1:-1]
        # share new values
        values = MCH.master_broadcasts_array_to_all(values)
        Niter += 1
    # adapt Niter for correct return
    Niter -= 1
    # share master current amp to all process
    current_amp = MCH.master_broadcasts_array_to_all(current_amp)
    return current_amp, Niter



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import numpy as np
import matplotlib.pyplot as plt


from .NRV_Class import NRV_class, load_any
from .NRV_Results import NRV_results
from .log_interface import rise_warning
from ..fmod.stimulus import stimulus

class sim_results(NRV_results):
    def __init__(self, context=None):
        super().__init__(context)

    def plot_stim(self, IDs=None, t_stop=None, N_pts=1000, ax=None, **fig_kwargs):
        """
        Plot one or several stimulis of the simulation extra-cellular context 
        """
        if 'extra_stim' not in self:
            rise_warning('No extracellular stimulation to be plotted')
        else:
            if IDs is None:
                IDs = [i for i in range(len(self['extra_stim']['stimuli']))]
            elif not np.iterable(IDs):
                IDs = [int(IDs)]
            for i in IDs:
                stim = load_any(self['extra_stim']['stimuli'][i])
                if t_stop is None:
                    t_stop = stim.t[-1]
                stim2 = stimulus()
                stim2.s = np.zeros(N_pts)
                stim2.t = np.linspace(0, t_stop, N_pts)
                stim2 += stim
                if ax is None:
                    plt.plot(stim2.t, stim2.s, **fig_kwargs)
                else:
                    ax.plot(stim2.t, stim2.s, **fig_kwargs)




class NRV_simulable(NRV_class):
    """
    
    """
    def __init__(
        self,
        t_sim=2e1,
        dt=0.001,
        record_V_mem=True,
        record_I_mem=False,
        record_I_ions=False,
        record_particles=False,
        record_g_mem=False,
        record_g_ions=False,
        loaded_footprints=None,
    ):
        super().__init__()
        self.t_sim = t_sim
        self.record_V_mem = record_V_mem
        self.record_I_mem = record_I_mem
        self.record_I_ions = record_I_ions
        self.record_particles = record_particles
        self.record_g_mem = record_g_mem
        self.record_g_ions = record_g_ions
        self.loaded_footprints = loaded_footprints

        self.dt = dt

    def extracel_status(self):
        if "extra_stim" in self.__dict__:
            if self.__dict__["extra_stim"] is not None:
                return True
        else:
            return False

    def intracel_status(self):
        if "intra_current_stim" in self.__dict__:
            if self.__dict__["intra_current_stim"] != [] and self.__dict__["intra_current_stim"] is not None:
                return True
        elif "intra_voltage_stim" in self.__dict__:
            if self.__dict__["intra_voltage_stim"] is not None:
                return True
        else:
            return False

    def rec_status(self):
        if "recorder" in self.__dict__:
            if self.__dict__["recorder"] is not None:
                return True
        else:
            return False

    def __call__(self, **kwds):
        return self.simulate(**kwds)

    def simulate(self, **kwargs) -> sim_results:
        self.set_parameters(**kwargs)
        context = self.save(
                save=False,
                extracel_context=self.extracel_status(),
                intracel_context=self.intracel_status(),
                rec_context=self.rec_status(),
            )
        results = sim_results(context)
        return results



from cmath import phase
import numpy as np

from ....backend.NRV_Class import NRV_class
from ....backend.log_interface import rise_error, rise_warning
from ....utils.units import mm
from ....backend.file_handler import rmv_ext
from .MshCreator import MshCreator, pi

ENT_DOM_offset = {
    "Volume": 0,
    "Surface": 1,
    "Outerbox": 0,
    "Nerve": 2,
    "Fascicle": 10,
    "Electrode": 100,
    "Axon": 1000,
}

ELEC_TYPES = ["CUFF MP", "CUFF", "LIFE"]


def is_NerveMshCreator(object):
    """
    check if an object is a NerveMshCreator, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a NerveMshCreator object
    """
    return isinstance(object, NerveMshCreator)


class NerveMshCreator(MshCreator):
    """
    Class allowing to generate Nerve shape 3D gmsh mesh with labeled physical domain
    Contains methodes dealing with the mesh geometries, physical domains and feilds
    Inherit from MshCreator class. see MshCreator for further detail
    """

    def __init__(
        self, Length=10000, Outer_D=5, Nerve_D=4000, y_c=0, z_c=0, ver_level=2,
    ):
        """
        initialisation of the NerveMshCreator

        Parameters
        ----------
        Length          : float
            Nerve length in um, by default 10000
        Outer_D         : float
            Outer box diameter in mm, by default 5
        Nerve_D         : float
            Nerve diameter in um, by default 4000
        y_c             : float
            y-axis position of the Nerve center, by default 0
        z_c             : float
            z-axis position of the Nerve center, by default 0
        ver_level       : int(0,1,2,3,4,5,99)
            verbosity level of gmsh (see MshCreator.set_verbosity), by default 2
        """
        super().__init__(D=3, ver_level=ver_level)
        self.L = Length
        self.y_c = y_c
        self.z_c = z_c
        self.surf_c = [self.L / 2]

        self.gnd_facet = {"outfacet": True, "lfacet": False, "rfacet": False}
        if Outer_D is None:
            self.is_outer = False
            self.gnd_facet["outfacet"] = False
            self.gnd_facet["rfacet"] = True
            self.Outer_D = 0
            self.Outer_entities = {"face": [], "volume": []}
        else:
            self.is_outer = True
            self.Outer_D = Outer_D * mm
            self.Outer_entities = {"face": [], "volume": []}

        self.Nerve_D = Nerve_D
        self.Nerve_entities = {"face": [], "volume": []}

        self.N_fascicle = 0
        self.fascicles = {}
        self.N_axon = 0
        self.axons = {}

        self.N_electrode = 0
        self.electrodes = {}

        self.default_res = {
            "Outerbox": self.Outer_D / 5,
            "Outerbox_tresholded": False,
            "Nerve": self.Nerve_D / 5,
            "Fascicle": 500,
            "Axon": 10,
            "Electrode": 1000,
        }
        self.Ox = None  # add an Ox line (need for thresholded resolution)

        self.res = self.default_res["Outerbox"]
        self.alpha_Outerboxres = 0.05

        self.is_geo = False
        self.is_dom = False
        self.is_refined = False

    def get_parameters(self):
        param = {}
        param["res"] = self.res
        param["L"] = self.L
        param["y_c"] = self.y_c
        param["z_c"] = self.z_c
        param["surf_c"] = self.surf_c
        param["Outer_D"] = self.Outer_D
        param["Outer_entities"] = self.Outer_entities
        param["Nerve_D"] = self.Nerve_D
        param["Nerve_entities"] = self.Nerve_entities
        param["N_fascicle"] = self.N_fascicle
        param["fascicles"] = self.fascicles
        param["N_axon"] = self.N_axon
        param["axons"] = self.axons
        param["N_electrode"] = self.N_electrode
        param["electrodes"] = self.electrodes
        return param

    def save(
        self,
        fname="nervemshcreator.json",
        save=True,
        mshfname=None,
        blacklist=[],
        **kwargs,
    ):
        """
        Return extracellular context as dictionary and eventually save it as json file
        NB: caution, first argument is fname to match with MshCreator.save arguments

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "extracel_context.json"

        Returns
        -------
        context_dic : dict
            dictionary containing all information
        """
        blacklist += [
            "model",
        ]
        if self.is_generated and save:
            if mshfname is None:
                mshfname = rmv_ext(fname) + ".msh"
            super().save(fname=mshfname, generate=False)
        fname = rmv_ext(fname) + ".json"
        return NRV_class.save(
            self, save=save, fname=fname, blacklist=blacklist, **kwargs
        )

    def load(self, data, mshfname=None, blacklist={}, **kwargs):
        """
        Generic loading method for NRV class instance

        Parameters
        ----------
        data : dict
            Dictionary containing the NRV object
        blacklist : dict, optional
            Dictionary containing the keys to be excluded from the load
        **kwargs : dict, optional
            Additional arguments to be passed to the load method of the NRV object
        """
        NRV_class.load(self, data, **kwargs)
        self.electrodes = {int(k): v for k, v in self.electrodes.items()}
        self.fascicles = {int(k): v for k, v in self.fascicles.items()}
        self.axons = {int(k): v for k, v in self.axons.items()}
        if self.is_generated and self.file != "":
            super().load(self.file)

    def compute_mesh(self):
        """
        Compute mesh geometry, domains and resolution and then generate the mesh
        """
        self.compute_geo()
        self.compute_domains()
        self.compute_res()
        self.generate()

    def set_gnd_facet(self, outfacet=None, lfacet=None, rfacet=None):
        """
        Set which of the outer facet should be the ground (element 1)

        Parameters
        ----------
        outfacet        : bool or None
            if true outer ring facet is included to the element 1, if None not modified
            by default None
        lfacet          : bool or None
            if true left facet is included to the element 1, if None not modified
            by default None
        rfacet          : bool or None
            if true right facet is included to the element 1, if None not modified
            by default None
        """
        if outfacet is not None:
            self.gnd_facet["outfacet"] = outfacet
        if lfacet is not None:
            self.gnd_facet["lfacet"] = lfacet
        if rfacet is not None:
            self.gnd_facet["rfacet"] = rfacet

    ###############################################################################################
    #####################################   geometry definition  ##################################
    ###############################################################################################

    def compute_geo(self):
        """
        Compute the mesh geometry

        """
        if not self.is_geo:
            if self.is_outer:
                self.add_cylinder(0, self.y_c, self.z_c, self.L, self.Outer_D / 2)
                if self.default_res["Outerbox_tresholded"]:
                    self.Ox = self.add_line((0, 0, 0), (self.L, 0, 0))
            self.add_cylinder(0, self.y_c, self.z_c, self.L, self.Nerve_D / 2)
            for i in self.fascicles:
                fascicle = self.fascicles[i]
                self.add_cylinder(
                    0, fascicle["y_c"], fascicle["z_c"], self.L, fascicle["D"] / 2
                )

            for i in self.axons:
                axon = self.axons[i]
                self.add_cylinder(0, axon["y_c"], axon["z_c"], self.L, axon["D"] / 2)
            for i in self.electrodes:
                electrode = self.electrodes[i]
                if "CUFF MP" in electrode["type"]:
                    self.add_CUFF_MP(ID=i, **electrode["kwargs"])
                elif "CUFF MEA" in electrode["type"]:
                    self.add_CUFF_MEA(ID=i, **electrode["kwargs"])
                elif "CUFF" in electrode["type"]:
                    self.add_CUFF(ID=i, **electrode["kwargs"])
                elif "LIFE" in electrode["type"]:
                    self.add_LIFE(ID=i, **electrode["kwargs"])

            self.fragment(verbose=False)
            self.is_geo = True

    def reshape_outerBox(self, Outer_D=None, res="default", tresholded_res=None):
        """
        Reshape the size of the FEM simulation outer box

        Parameters
        ----------
        outer_D : float
            FEM simulation outer box diameter, in mm, WARNING, this is the only parameter in mm !
        """
        if Outer_D is not None:
            """
            ## See how if it should be added
            if self.default_res["Outerbox"] == self.Outer_D/5:
                self.default_res["Outerbox"] = Outer_D/5
            """
            self.Outer_D = Outer_D * mm

        if tresholded_res is not None:
            self.default_res["Outerbox_tresholded"] = tresholded_res

        if not res == "default":
            self.default_res["Outerbox"] = res

        if self.default_res["Outerbox"] > self.Outer_D / 3:
            self.default_res["Outerbox"] = self.Outer_D / 3

    def reshape_nerve(
        self, Nerve_D=None, Length=None, y_c=None, z_c=None, res="default"
    ):
        """
        Reshape the nerve of the FEM simulation

        Parameters
        ----------
        Nerve_D                 : float
            Nerve diameter, in um
        Length                  : float
            Nerve length, in um
        y_c                     : float
            Nerve center y-coordinate in um, 0 by default
        z_c                     : float
            Nerve z-coordinate center in um, 0 by default
        """

        if Length is not None:
            self.L = Length
        if Nerve_D is not None:
            if self.default_res["Nerve"] == self.Nerve_D / 5:
                self.default_res["Nerve"] = Nerve_D / 5
            self.Nerve_D = Nerve_D
        if y_c is not None:
            self.y_c = y_c
        if z_c is not None:
            self.z_c = z_c

        if not res == "default":
            self.default_res["Nerve"] = res

        if self.default_res["Nerve"] > self.Nerve_D / 5:
            self.default_res["Nerve"] = self.Nerve_D / 5

    def reshape_fascicle(self, D, y_c=0, z_c=0, ID=None, res="default"):
        """
        Reshape a fascicle of the FEM simulation

        Parameters
        ----------
        Fascicle_D  : float
            Fascicle diameter, in um
        y_c         : float
            Fascicle center y-coodinate in um, 0 by default
        z_c         : float
            Fascicle center y-coodinate in um, 0 by default
        ID          : int
            If the simulation contains more than one fascicles, ID number of the fascicle to reshape as in COMSOL
        """
        if ID not in self.fascicles:
            if ID is None:
                ID = 0
                while ID in self.fascicles:
                    ID += 1
            self.N_fascicle += 1

        if res == "default":
            res = self.default_res["Fascicle"]
        if D / 5 < res:
            res = D / 5

        self.fascicles[ID] = {
            "y_c": y_c,
            "z_c": z_c,
            "D": D,
            "res": res,
            "face": None,
            "volume": None,
        }

    def remove_fascicles(self, ID=None):
        """
        remove a fascicle of the FEM simulation

        Parameters
        ----------
        ID          : int, None
            ID number of the fascicle to remove, if None, remove all fascicles, by default None
        """
        if ID is None:
            self.fascicles = {}
            self.N_fascicle = 0
        elif ID in self.fascicles:
            del self.fascicles[ID]
            self.N_fascicle -= 1

    def reshape_axon(self, D, y_c=0, z_c=0, ID=None, res="default"):
        """
        Reshape a axon of the FEM simulation

        Parameters
        ----------
        Fascicle_D  : float
            Fascicle diameter, in um
        y_c         : float
            Fascicle center y-coodinate in um, 0 by default
        z_c         : float
            Fascicle center y-coodinate in um, 0 by default
        ID          : int
            If the simulation contains more than one fascicles, ID number of the fascicle to reshape as in COMSOL
        """
        if ID not in self.axons:
            if ID is None:
                ID = 0
                while ID in self.axons:
                    ID += 1
            self.N_axon += 1

        if res == "default":
            res = self.default_res["Axon"]
        if D / 3 < res:
            res = D / 3

        self.axons[ID] = {
            "y_c": y_c,
            "z_c": z_c,
            "D": D,
            "res": res,
            "face": None,
            "volume": None,
        }

    def add_electrode(self, elec_type, ID=None, res="default", **kwargs):
        """ """
        if elec_type not in ELEC_TYPES + ["CUFF MEA"]:
            rise_warning(
                elec_type
                + " not implemented, electrode will not be added\nList of Electrode types: "
                + str(ELEC_TYPES)
            )
        if ID not in self.electrodes:
            if ID is None:
                ID = 0
                while ID in self.electrodes:
                    if (
                        "CUFF MEA" in self.electrodes[ID]["type"]
                        or "CUFF MP" in self.electrodes[ID]["type"]
                    ):
                        ID += self.electrodes[ID]["kwargs"]["N"]
                    else:
                        ID += 1
            self.N_electrode += 1

        if res == "default":
            res = self.default_res["Electrode"]

        if "LIFE" in elec_type:
            if "D" in kwargs:
                D = kwargs["D"]
            else:
                D = 25
            if D / 3 < res:
                res = D / 3

        self.electrodes[ID] = {"type": elec_type, "res": res, "kwargs": kwargs}

    ####################################################################################################
    #####################################   domains definition  ########################################
    ####################################################################################################

    def compute_domains(self):
        if not self.is_geo:
            rise_error("compute geometry before domain")
        elif not self.is_dom:
            self.__link_entity_domains(2)
            self.__link_entity_domains(3)
            self.compute_entity_domain()
            self.is_dom = True

    def __is_outerbox(self, dx, dy, dz, com, dim_key):
        """
        Internal use only: check if volume is the box or face is external face of box

        Parameters
        ----------
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox

        """
        outer_ring_test = (
            np.allclose([dx, dy, dz], [self.L, self.Outer_D, self.Outer_D])
            and self.is_outer
        )
        if dim_key == "volume":
            return outer_ring_test
        outer_ring_test = outer_ring_test and self.gnd_facet["outfacet"]
        left_facet_test = np.allclose(com[0], [0]) and self.gnd_facet["lfacet"]
        rigth_facet_test = np.allclose(com[0], [self.L]) and self.gnd_facet["rfacet"]
        return outer_ring_test or left_facet_test or rigth_facet_test

    def __is_nerve(self, dx, dy, dz):
        """
        Internal use only: check if volume is nerve or face is external face of nerve

        Parameters
        ----------
        dx          : float
            length along x of the tested entity
        dy          : float
            length along y of the tested entity
        dz          : float
            length along z of the tested entity
        """
        return np.allclose([dy, dz], [self.Nerve_D, self.Nerve_D]) and not np.isclose(
            dx, 0
        )

    def __is_fascicle(self, ID, dx, dy, dz, com, dim_key):
        """
        Internal use only: check if volume is fascicle ID or face is external face of fascicle ID

        Parameters
        ----------
        ID          :int
            ID of fascicle to test
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        dim_key     :"face" or "volume"
            element type
        """
        fascicle = self.fascicles[ID]
        # Already computed
        status_test = fascicle[dim_key] is None
        # test good diameter
        size_test = np.allclose([dx, dy, dz], [self.L, fascicle["D"], fascicle["D"]])
        # test center of mass in fascicle
        com_test = np.allclose(
            com,
            (self.L / 2, fascicle["y_c"], fascicle["z_c"]),
            rtol=1,
            atol=fascicle["D"] / 2,
        )
        return status_test and size_test and com_test

    def __is_axon(self, ID, dx, dy, dz, com, dim_key):
        """
        Internal use only: check if volume is axon ID or face is external face of axon ID

        Parameters
        ----------
        ID          :int
            ID of axon to test
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        dim_key     :"face" or "volume"
            element type
        """
        axon = self.axons[ID]
        # Already computed
        status_test = axon[dim_key] is None
        # test good diameter
        size_test = np.allclose([dx, dy, dz], [self.L, axon["D"], axon["D"]])
        # test center of mass in axon
        com_test = np.allclose(
            com, (self.L / 2, axon["y_c"], axon["z_c"]), rtol=1, atol=axon["D"] / 2
        )
        return status_test and size_test and com_test

    def __is_CUFF_MP_electrode(self, ID, rc, dx, teta, com, dim_key):
        """
        Internal use only: check if volume is electrode ID or face is external face of fascicle ID

        Parameters
        ----------
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        """
        elec_kwargs = self.electrodes[ID]["kwargs"]
        if self.electrodes[ID]["type"] != "CUFF MP":
            return False

        # test good length
        size_test = np.allclose([dx], [elec_kwargs["contact_length"]])
        # test good x-location
        com_test = np.isclose(com[0], elec_kwargs["x_c"])
        # test good polar coordinate
        if elec_kwargs["is_volume"] or dim_key == "volume":
            N = elec_kwargs["N"]
            tetas = [teta for i in range(N)]
            angle_test = np.isclose(tetas, self.electrodes[ID]["angles"]).any()
            radius_test = rc > self.Nerve_D / 2
            polar_test = angle_test and radius_test
        else:
            polar_test = rc <= self.Nerve_D / 2

        return size_test and com_test and polar_test

    def __is_CUFF_electrode(self, ID, dx, dy, dz, com, dim_key):
        """
        Internal use only: check if volume is electrode ID or face is external face of fascicle ID

        Parameters
        ----------
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        """
        elec_kwargs = self.electrodes[ID]["kwargs"]
        if self.electrodes[ID]["type"] != "CUFF":
            return False

        # test good diameter
        Syz = self.Nerve_D
        if elec_kwargs["is_volume"] or dim_key == "volume":
            Syz += 2 * elec_kwargs["contact_thickness"]
        size_test = np.allclose([dx, dy, dz], [elec_kwargs["contact_length"], Syz, Syz])
        # test center of mass in CUFF
        com_test = np.allclose(com, (elec_kwargs["x_c"], self.y_c, self.z_c))
        return size_test and com_test

    def __is_LIFE_electrode(self, ID, dx, dy, dz, com):
        """
        Internal use only: check if volume is electrode ID or face is external face of fascicle ID

        Parameters
        ----------
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        """

        elec_kwargs = self.electrodes[ID]["kwargs"]
        if self.electrodes[ID]["type"] != "LIFE":
            return False
        # test good diameter
        size_test = np.allclose(
            [dx, dy, dz], [elec_kwargs["length"], elec_kwargs["D"], elec_kwargs["D"]]
        )
        # test center of mass in LIFE
        com_test = np.allclose(
            com,
            (elec_kwargs["x_c"], elec_kwargs["y_c"], elec_kwargs["z_c"]),
            atol=elec_kwargs["D"] / 2,
        )
        return size_test and com_test

    def __link_entity_domains(self, dim):
        """
        Internal use only: link all entities from

        Parameters
        ----------
        dx          : float
            length along x of the entity boundbox
        dy          : float
            length along y of the entity boundbox
        dz          : float
            length along z of the entity boundbox
        com         : tupple(float)
            entity Center of Mass
        """
        if dim == 2:
            entities, ent_com, ent_bd = self.get_faces(com=True, bd=True)
            key = "face"
            offset = 1
        else:
            entities, ent_com, ent_bd = self.get_volumes(com=True, bd=True)
            key = "volume"
            offset = 0

        for i in range(len(entities)):
            bd_x = abs(ent_bd[i][3] - ent_bd[i][0])
            bd_y = abs(ent_bd[i][4] - ent_bd[i][1])
            bd_z = abs(ent_bd[i][5] - ent_bd[i][2])
            if self.__is_outerbox(bd_x, bd_y, bd_z, ent_com[i], key):
                self.Outer_entities[key] += [entities[i][1]]

            elif self.__is_nerve(bd_x, bd_y, bd_z):
                self.Nerve_entities[key] += [entities[i][1]]

            for j in self.fascicles:
                if self.__is_fascicle(j, bd_x, bd_y, bd_z, ent_com[i], key):
                    self.fascicles[j][key] = entities[i][1]

            for j in self.axons:
                if self.__is_axon(j, bd_x, bd_y, bd_z, ent_com[i], key):
                    self.axons[j][key] = entities[i][1]

            for j in self.electrodes:
                r_c = (
                    (ent_com[i][1] - self.y_c) ** 2 + (ent_com[i][2] - self.z_c) ** 2
                ) ** 0.5
                teta = phase(
                    complex(ent_com[i][2] - self.z_c, ent_com[i][1] - self.y_c)
                ) % (2 * pi)
                if self.__is_CUFF_MP_electrode(j, r_c, bd_x, teta, ent_com[i], key):
                    N = self.electrodes[j]["kwargs"]["N"]
                    ID_EA = round(teta * N / (2 * pi)) % N
                    if dim == 3:
                        if self.electrodes[j][key][ID_EA] is None:
                            self.electrodes[j][key][ID_EA] = entities[i][1]
                        else:
                            rise_warning(
                                "two volumes can be same electrode only the first one is kept"
                            )
                    else:
                        if self.electrodes[j][key][ID_EA] is None:
                            self.electrodes[j][key][ID_EA] = entities[i][1]
                        else:
                            self.electrodes[j][key][ID_EA] = [
                                entities[i][1],
                                self.electrodes[j][key][ID_EA],
                            ]

                elif self.__is_CUFF_electrode(j, bd_x, bd_y, bd_z, ent_com[i], key):
                    self.electrodes[j][key] = entities[i][1]
                    if not self.electrodes[j]["kwargs"]["is_volume"] and key == "face":
                        self.Nerve_entities[key].remove(entities[i][1])

                elif self.__is_LIFE_electrode(j, bd_x, bd_y, bd_z, ent_com[i]):
                    self.electrodes[j][key] = entities[i][1]

    def compute_entity_domain(self):
        """
        link all geometrical entities to corresponding physical domains
        """
        # Outer box domain
        if self.is_outer:
            self.add_domains(obj_IDs=self.Outer_entities["volume"], phys_ID=0, dim=3)
            self.add_domains(obj_IDs=self.Outer_entities["face"], phys_ID=1, dim=2)
        else:
            self.add_domains(obj_IDs=self.Outer_entities["face"], phys_ID=1, dim=2)

        # nerve domain
        self.add_domains(obj_IDs=self.Nerve_entities["volume"], phys_ID=2, dim=3)
        self.add_domains(obj_IDs=self.Nerve_entities["face"], phys_ID=3, dim=2)

        for j in self.fascicles:
            id_ph = ENT_DOM_offset["Fascicle"] + (2 * j)
            if id_ph < ENT_DOM_offset["Electrode"]:
                self.add_domains(
                    obj_IDs=self.fascicles[j]["volume"], phys_ID=id_ph, dim=3
                )
                id_ph += ENT_DOM_offset["Surface"]
                self.add_domains(
                    obj_IDs=self.fascicles[j]["face"], phys_ID=id_ph, dim=2
                )
            else:
                rise_warning("Too much Fascicles: " + str(j) + " not added")

        for j in self.axons:
            id_ph = ENT_DOM_offset["Axon"] + (2 * j)
            self.add_domains(obj_IDs=self.axons[j]["volume"], phys_ID=id_ph, dim=3)
            id_ph += ENT_DOM_offset["Surface"]
            self.add_domains(obj_IDs=self.axons[j]["face"], phys_ID=id_ph, dim=2)

        for j in self.electrodes:
            id_ph = ENT_DOM_offset["Electrode"] + (2 * j)
            if id_ph < ENT_DOM_offset["Axon"]:
                if "CUFF MP" in self.electrodes[j]["type"]:
                    for ID_EA in range(self.electrodes[j]["kwargs"]["N"]):
                        if self.electrodes[j]["kwargs"]["is_volume"]:
                            self.add_domains(
                                obj_IDs=self.electrodes[j]["volume"][ID_EA],
                                phys_ID=id_ph,
                                dim=3,
                            )
                        id_ph += ENT_DOM_offset["Surface"]
                        self.add_domains(
                            obj_IDs=self.electrodes[j]["face"][ID_EA],
                            phys_ID=id_ph,
                            dim=2,
                        )
                        id_ph += 1
                else:
                    if self.electrodes[j]["kwargs"]["is_volume"]:
                        self.add_domains(
                            obj_IDs=self.electrodes[j]["volume"], phys_ID=id_ph, dim=3
                        )
                    id_ph += ENT_DOM_offset["Surface"]
                    self.add_domains(
                        obj_IDs=self.electrodes[j]["face"], phys_ID=id_ph, dim=2
                    )
            else:
                rise_warning("Too much Electrodes: " + str(j) + " not added")

    ####################################################################################################
    ###################################   field (res) definition  ######################################
    ####################################################################################################

    def compute_res(self):
        if not self.is_dom and self.is_geo:
            rise_error("compute geometry and domain before resolution")
        elif not self.is_refined:
            self.res = max(self.default_res.values())
            fields = []

            # Outerbox field
            if self.is_outer:
                fields += self.__refine_Outer_box(alpha=self.alpha_Outerboxres)
            # Nerve field
            fields += [
                self.refine_entities(
                    ent_ID=self.Nerve_entities["volume"],
                    res_in=self.default_res["Nerve"],
                    dim=3,
                    res_out=None,
                    IncludeBoundary=True,
                )
            ]
            # Facsicle fields
            for j in self.fascicles:
                fascicle = self.fascicles[j]
                fields += [
                    self.refine_entities(
                        ent_ID=fascicle["volume"],
                        res_in=fascicle["res"],
                        dim=3,
                        res_out=None,
                        IncludeBoundary=True,
                    )
                ]
            # Axon fields
            for j in self.axons:
                axon = self.axons[j]
                fields += [
                    self.refine_entities(
                        ent_ID=axon["volume"],
                        res_in=axon["res"],
                        dim=3,
                        res_out=None,
                        IncludeBoundary=True,
                    )
                ]
            # Electrodes fields
            for j in self.electrodes:
                electrode = self.electrodes[j]

                if electrode["type"] in ["CUFF MEA", "CUFF MP"]:
                    for ID_EA in range(electrode["kwargs"]["N"]):
                        fields += [
                            self.refine_entities(
                                ent_ID=electrode["volume"][ID_EA],
                                res_in=electrode["res"],
                                dim=3,
                                res_out=None,
                                IncludeBoundary=True,
                            )
                        ]
                else:
                    fields += [
                        self.refine_entities(
                            ent_ID=electrode["volume"],
                            res_in=electrode["res"],
                            dim=3,
                            res_out=None,
                            IncludeBoundary=True,
                        )
                    ]

            self.refine_min(fields)
            self.is_refined = True

    def __refine_Outer_box(self, alpha=0.1):
        """ """
        if not self.default_res["Outerbox_tresholded"]:
            field = [
                self.refine_entities(
                    ent_ID=self.Outer_entities["volume"],
                    res_in=self.default_res["Outerbox"],
                    dim=3,
                    res_out=None,
                    IncludeBoundary=True,
                )
            ]
        else:
            dmin = self.Nerve_D
            dmax = dmin + alpha * self.Outer_D
            field = [
                self.refine_threshold(
                    ent_ID=self.Ox,
                    dim=1,
                    res_min=self.default_res["Nerve"],
                    res_max=self.default_res["Outerbox"],
                    dist_min=self.Nerve_D,
                    dist_max=dmax,
                )
            ]
        return field

    ####################################################################################################
    ###################################   electrodes definition  #######################################
    ####################################################################################################

    def add_LIFE(
        self, ID=None, x_c=0, y_c=0, z_c=0, length=1000, D=25, is_volume=False
    ):
        """
        Add LIFE electrode to the mesh

        Parameters
        ----------
        ID          :int
            electrod ID, ,by defalt None
        x_c         :float
            x-position of the LIFE center in um, by default 0
        y_c         :float
            y-position of the LIFE center in um, by default 0
        z_c         :float
            z-position of the LIFE center in um, by default 0
        length      :float
            length of the LIFE electrod in um, by default 1000
        D           :float
            diameter of the LIFE electrod in um, by default 25
        is_volume   : bool
            if True in cylinder of the LIFE is kept on the mesh

        """

        if ID is not None:
            self.electrodes[ID]["volume"] = []
            self.electrodes[ID]["face"] = []
            self.electrodes[ID]["kwargs"]["is_volume"] = is_volume
        x_active = x_c - length / 2
        y_active = y_c
        z_active = z_c
        self.add_cylinder(x_active, y_active, z_active, length, D / 2)

    def add_CUFF(
        self,
        ID=None,
        x_c=0,
        contact_length=100,
        is_volume=True,
        contact_thickness=None,
        insulator=True,
        insulator_thickness=None,
        insulator_length=None,
        insulator_offset=0,
    ):
        """
        Add CUFF electrode to the mesh

        Parameters
        ----------
        ID                      :int
            if not none and ID exist, change electrod ID,by defalt None
        x_c                     :float
            x-position of the CUFF center in um, by default 0
            length of the CUFF electrod in um, by default 100
        contact_length       :float
            length along x of the contact site in um, by default 100
        is_volume   : bool
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness       :float
            thickness of the contact site in um, by default 5
        insulator                :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness     :float
            thickness of the insulator ring in um, by default 20
        insulator_length        :float
            length along x of the insulator ring in um, by default 1000
        """

        if contact_thickness is None:
            contact_thickness = 0.1 * (self.Outer_D - self.Nerve_D) / 2
            if contact_thickness < 0:
                contact_thickness = self.Nerve_D * 0.1
        if insulator_thickness is None:
            insulator_thickness = min(
                5 * contact_thickness, 0.4 * (self.Outer_D - self.Nerve_D) / 2
            )
        if insulator_length is None:
            insulator_length = 2 * contact_length
        # self.reshape_outerBox(tresholded_res=True)

        if ID is not None:
            self.electrodes[ID]["kwargs"]["contact_length"] = contact_length
            self.electrodes[ID]["kwargs"]["contact_thickness"] = contact_thickness
            self.electrodes[ID]["kwargs"]["is_volume"] = is_volume
            self.electrodes[ID]["kwargs"]["insulator_offset"] = insulator_offset
            self.electrodes[ID]["volume"] = []
            self.electrodes[ID]["face"] = []
            if contact_length / 3 < self.electrodes[ID]["res"]:
                self.electrodes[ID]["res"] = contact_length / 2

        x_active = x_c - contact_length / 2
        y_active = self.y_c
        z_active = self.z_c
        cyl_act = self.add_cylinder(
            x_active,
            y_active,
            z_active,
            contact_length,
            self.Nerve_D / 2 + contact_thickness,
        )
        cyl_ner2 = self.add_cylinder(
            x_active, y_active, z_active, contact_length, self.Nerve_D / 2
        )
        self.model.occ.cut([(3, cyl_act)], [(3, cyl_ner2)])

        if insulator:
            x_insulator = x_c + insulator_offset - insulator_length / 2
            y_insulator = self.y_c
            z_insulator = self.z_c

            cyl_ina = self.add_cylinder(
                x_insulator,
                y_insulator,
                z_insulator,
                insulator_length,
                self.Nerve_D / 2 + insulator_thickness,
            )
            cyl_ner = self.add_cylinder(
                x_insulator,
                y_insulator,
                z_insulator,
                insulator_length,
                self.Nerve_D / 2,
            )
            self.model.occ.cut([(3, cyl_ina)], [(3, cyl_ner), (3, cyl_act)])

    def add_CUFF_MP(
        self,
        ID=None,
        N=4,
        x_c=0,
        contact_width=None,
        contact_length=100,
        is_volume=True,
        contact_thickness=None,
        insulator=True,
        insulator_thickness=None,
        insulator_length=None,
        insulator_offset=0,
    ):
        """
        Add MultiPolar CUFF electrodes to the mesh

        Parameters
        ----------
        ID                      :int
            if not none and ID exist, change electrod ID,by defalt None
        N                       :int
            Number of active site, by default 4
        x_c                     :float
            x-position of the CUFF center in um, by default 0
            length of the CUFF electrod in um, by default 100
        contact_width           :float or None
            width of the active sites around the nerve, if None set to cover 4/5 of the perimeter
            with active sites
        contact_length          :float
            length along x of the contact site in um, by default 100
        is_volume   : bool
            if True the contact is kept on the mesh as a volume, by default True
        contact_thickness       :float
            thickness of the contact site in um, by default 5
        insulator                :bool
            remove insulator ring from the mesh (no conductivity), by default True
        insulator_thickness     :float
            thickness of the insulator ring in um, by default 20
        insulator_length        :float
            length along x of the insulator ring in um, by default 1000
        """
        if contact_width is None:
            contact_width = pi * self.Nerve_D * 4 / (5 * N)

        if contact_thickness is None:
            contact_thickness = 0.1 * (self.Outer_D - self.Nerve_D) / 2
            if contact_thickness < 0:
                contact_thickness = self.Nerve_D * 0.1
        if insulator_thickness is None:
            insulator_thickness = min(
                5 * contact_thickness, 0.4 * (self.Outer_D - self.Nerve_D) / 2
            )
        if insulator_length is None:
            insulator_length = 2 * contact_length

        if ID is not None:
            self.electrodes[ID]["kwargs"]["contact_width"] = contact_width
            self.electrodes[ID]["kwargs"]["contact_length"] = contact_length
            self.electrodes[ID]["kwargs"]["contact_thickness"] = contact_thickness
            self.electrodes[ID]["kwargs"]["insulator_thickness"] = insulator_thickness
            self.electrodes[ID]["kwargs"]["insulator_length"] = insulator_length
            self.electrodes[ID]["kwargs"]["insulator_offset"] = insulator_offset
            self.electrodes[ID]["kwargs"]["is_volume"] = is_volume
            self.electrodes[ID]["volume"] = [None for i in range(N)]
            self.electrodes[ID]["face"] = [None for i in range(N)]
            if contact_length / 3 < self.electrodes[ID]["res"]:
                self.electrodes[ID]["res"] = contact_length / 2

        angles = []
        bar_fus = []

        x_active = x_c - contact_length / 2
        y_active = self.y_c - contact_width / 2
        y_c = self.y_c
        z_c = self.z_c

        if insulator:
            if insulator_thickness is None:
                insulator_thickness = contact_thickness * 3
            if insulator_length is None:
                insulator_length = contact_length * 2

            x_insulator = x_c + insulator_offset - insulator_length / 2

            cyl1 = self.add_cylinder(
                x_insulator,
                y_c,
                z_c,
                insulator_length,
                self.Nerve_D / 2 + insulator_thickness,
            )
            cyl2 = self.add_cylinder(
                x_insulator, y_c, z_c, insulator_length, self.Nerve_D / 2
            )
            self.model.occ.cut([(3, cyl1)], [(3, cyl2)])

        for i in range(N):
            bar = self.add_box(
                x_active,
                y_active,
                z_c,
                contact_length,
                contact_width,
                contact_thickness + self.Nerve_D / 2,
            )
            angles += [(2 * pi * i) / (N)]
            self.rotate(bar, angles[-1], x_c, y_c, z_c, ax=1)
            bar_fus += [(3, bar)]

        if ID is not None:
            self.electrodes[ID]["angles"] = angles

        cyl = self.add_cylinder(0, self.y_c, self.z_c, self.L, self.Nerve_D / 2)
        self.model.occ.cut(bar_fus, [(3, cyl)])

    def add_CUFF_MEA(
        self,
        ID=None,
        N=4,
        x_c=0,
        y_c=0,
        z_c=0,
        size=None,
        thickness=100,
        inactive=True,
        inactive_th=None,
        inactive_L=None,
    ):
        """ """
        rise_warning("add_CUFF_MEA not updated use add_CUFF_MP instead")

        if size is None:
            s = pi * self.Nerve_D * 4 / (5 * N)
            size = (s, s)
        elif not np.iterable(size):
            size = (size, size)
        elif len(size) != 2:
            size = (size[0], size[0])

        self.add_CUFF_MP(
            ID=ID,
            N=N,
            x_c=x_c,
            contact_width=size[1],
            contact_length=size[0],
            is_volume=True,
            contact_thickness=thickness,
            insulator=inactive,
            insulator_thickness=inactive_th,
            insulator_length=inactive_L,
        )
        if ID is not None:
            self.electrodes[ID]["type"] = "CUFF MP"



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from .NRV_Class import NRV_class, abstractmethod, is_NRV_class


class NRV_results(NRV_class, dict):
    """
    Results class for NRV
    """
    @abstractmethod
    def __init__(self, context=None):
        super().__init__()
        if context is None:
            context = {}
        elif is_NRV_class(context):
            context.save(save=False)

        if "nrv_type" in context:
            context["result_type"] = context.pop("nrv_type")
        self.update(context)
        self.__sync()


    def save(self, save=False, fname="nrv_save.json", blacklist=[], **kwargs):
        self.__sync()
        return super().save(save, fname, blacklist, **kwargs)


    def load(self, data, blacklist=[], **kwargs):
        super().load(data, blacklist, **kwargs)
        self.__sync()


    def __setitem__(self, key, value):
        if not key == "nrv_type":
            self.__dict__[key] = value
        super().__setitem__(key, value)

    def __delitem__(self, key):
        if not key == "nrv_type":
            del self.__dict__[key]
        super().__delitem__(key)

    def update(self, __m, **kwargs) -> None:
        """
        overload of dict update method to update both attibute and items
        """
        self.__dict__.update(__m, **kwargs)
        super().update(__m, **kwargs)
    
    def __sync(self):
        self.update(self.__dict__)
        self.pop('__NRVObject__')



"""
NRV-Units
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""

###################
## defaulf units ##
###################
mV = 1  # default unit for Voltage is mV, as neuron
ms = 1  # default unit for time is ms, as neuron
um = 1  # default unit for length is um, as neuron
uA = 1  # default unit for current is uA, UNLESS NEURON who is by default in nA
S = 1  # default unit for conductance is S
kHz = 1  # default unit for frequency is kHz
uF = 1  # default unit for capacitance is uF

######################
## voltage prefixes ##
######################
uV = 1e-3 * mV
V = 1e3 * mV

#############################
## time prefixes and units ##
#############################
us = 1e-3 * ms
s = sec = 1e3 * ms
minute = 60 * sec
hour = 60 * minute
day = 24 * hour

########################
## frequency prefixes ##
########################

Hz = 1e-3 * kHz
MHz = 1e3 * kHz
GHz = 1e6 * kHz

#####################
## length prefixes ##
#####################
mm = 1e3 * um
cm = 1e1 * mm
dm = 1e1 * cm
m = 1e1 * dm

######################
## current prefixes ##
######################
nA = 1e-3 * uA
mA = 1e3 * uA
A = 1e3 * mA

###########################
## conductance prefixes ##
###########################
kS = 1e3 * S
mS = 1e-3 * S
uS = 1e-6 * S

###########################
## capacitance prefixes ##
###########################
F = 1e6 * uF
mF = 1e3 * uF
pF = 1e-3 * uF
nF = 1e-6 * uF

#############################
####  Usefull functions ####
#############################


def sci_round(number, digits=3):
    power = "{:e}".format(number).split("e")[1]
    return round(number, -(int(power) - digits))



"""
NRV-nrv_function
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from abc import abstractmethod

import numpy as np
from scipy.interpolate import CubicHermiteSpline, interp1d
from scipy.special import erf

from ..backend.file_handler import json_dump, json_load
from ..backend.log_interface import rise_warning
from ..backend.NRV_Class import NRV_class

#############################
## sigma functions classes ##
#############################
spy_interp1D_kind = [
    "linear",
    "nearest",
    "nearest-up",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "previous",
    "next",
]


class nrv_function(NRV_class):
    """
    Class containg all comon method of fonction used in nrv
    """

    @abstractmethod
    def __init__(self):
        super().__init__()
        self.f_type = "nrv_function"
        self.dim = 0

    def __call__(self, *X):
        return self.call_method(*X)

    @staticmethod
    def call_method(self, X):
        return X

    def __compatible(self, b):
        if not callable(b):
            return True
        elif "dim" in self.__dir__() and "dim" in b.__dir__():
            if self.dim == b.dim:
                return True
        else:
            rise_warning("Type not compatible for operation", type(self), type(b))
            return False

    def __neg__(self):
        c = eval(self.f_type)()
        c.call_method = lambda *X: -self(*X)
        return c

    def __abs__(self):
        c = eval(self.f_type)()
        c.call_method = lambda *X: abs(self(*X))
        return c

    def __add__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) + b(*X)
            else:
                c.call_method = lambda *X: self(*X) + b
            return c

    def __radd__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b * (X) + self(*X)
            else:
                c.call_method = lambda *X: b + self(*X)
            return c

    def __sub__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) - b(*X)
            else:
                c.call_method = lambda *X: self(*X) - b
            return c

    def __rsub__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b(*X) - self(*X)
            else:
                c.call_method = lambda *X: b - self(*X)
            return c

    def __mul__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: self(*X) * b(*X)
            else:
                c.call_method = lambda *X: self(*X) * b
            return c

    def __rmul__(self, b):
        if self.__compatible(b):
            c = eval(self.f_type)()
            if callable(b):
                c.call_method = lambda *X: b(*X) * self(*X)
            else:
                c.call_method = lambda *X: b * self(*X)
            return c


###############################################################
####################### 1D functions ##########################
###############################################################


class function_1D(nrv_function):
    """
    class containing function from IR to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_1D"
        self.dim = 1

    # @staticmethod
    def call_method(self, X):
        # Composition with N dim function
        if isinstance(X, nrv_function):
            c_type = "function_" + str(X.dim) + "D"
            c = eval(c_type)()
            c.call_method = lambda *x: self(X(*x))
            return c
        else:
            return None


class gaussian(function_1D):
    def __init__(self, mu=0, sigma=1):
        """
        gaussian function define as:
        f(x) = e^{-\frac{(x-\mu)^2}{2*\sigma^2}}
        """
        super().__init__()
        self.mu = mu
        self.sigma = sigma

    def call_method(self, X):
        res = super().call_method(X)
        if res is None:
            res = np.exp(
                -(((X - self.mu) / self.sigma) ** 2) / 2
            )  # /(self.sigma*(2*np.pi)**0.5)
        return res


class gate(function_1D):
    def __init__(self, mu=0, sigma=1, kind="Rational", N=None):
        """ """
        super().__init__()
        # self.kind =
        self.mu = mu
        self.sigma = sigma
        self.N = N
        self.kind = kind

    def call_method(self, X):
        res = super().call_method(X)
        if res is None:
            X_eff = (X - self.mu) / self.sigma
            if self.N is None:
                res = (np.sign(X_eff + 0.5) - np.sign(X_eff - 0.5)) / 2
            elif self.kind.lower() == "rational":
                res = 1 / ((2 * X_eff) ** (2 * self.N) + 1)
            else:
                res = (
                    erf((X_eff + (0.5 * self.sigma) / self.mu) / self.N)
                    - erf((X_eff - (0.5 * self.sigma) / self.mu) / self.N)
                ) / 2
        return res


###############################################################
####################### 2D functions ##########################
###############################################################


class function_2D(nrv_function):
    """
    class containing function from IR^2 to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_2D"
        self.dim = 2

    @staticmethod
    def call_method(self, X):
        return 1

class  ackley(function_2D):
    """
    Bi-dimensional Ackley function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array)-> np.array:
        return -20*np.exp(-0.2*np.sqrt(0.5*(x**2+y**2))) \
            - np.exp(0.5*(np.cos(2*np.pi*x)+np.cos(2*np.pi*y))) \
            + np.e + 20

class  beale(function_2D):
    """
    Bi-dimensional Beale function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array)-> np.array:
        return (1.5 - x + x*y)**2\
            + (2.25 - x + x*y**2)**2\
            + (2.625 - x + x*y**3)**2

class  goldstein_price(function_2D):
    """
    Bi-dimensional Goldstein function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array)-> np.array:
        return (1 + (x + y + 1)**2 * (19 - 14*x + 3*x**2 - 14*y + 6*x*y + 3*y**2))\
            * (30 + (2*x - 3*y)**2 *(18 - 32*x + 12*x**2 + 48*y - 36*x*y + 27*y**2))

class  booth(function_2D):
    """
    Bi-dimensional Booth function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array)-> np.array:
        return (x + 2*y - 7)**2 + (2*x + y - 5)**2

class  bukin6(function_2D):
    """
    Bi-dimensional Booth function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, x: np.array, y: np.array)-> np.array:
        return 100*np.sqrt(np.abs(y - 0.01*x**2)) + 0.01*np.abs(x+10)


###############################################################
####################### ND functions ##########################
###############################################################
class function_ND(nrv_function):
    """
    class containing function from IR^n to IR
    Such function can be call either on 1 value or on a ndarray (and applied on each value)
    """

    def __init__(self):
        super().__init__()
        self.f_type = "function_ND"
        self.dim = "N"

    @staticmethod
    def call_method(self, *X):
        return 1

class Id(function_ND):
    """

    """
    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array)-> np.array:
        if len(x)==1 and np.iterable(x[0]):
            x = x[0]
        return x

class  rosenbock(function_ND):
    """
        Multi-dimensional Rosenbock function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array)-> np.array:
        if len(x)==1 and np.iterable(x[0]):
            x = x[0]
        result = 0
        for i in range(1, len(x)):
            result += 100*(x[i] - x[i-1]**2)**2 + (1-x[i-1])**2
        return result

class  rastrigin(function_ND):
    """
    Multi-dimensional Rastrigin function
    """
    def __init__(self):
        super().__init__()

    def call_method(self, *x: np.array)-> np.array:
        if len(x)==1 and np.iterable(x[0]):
            x = x[0]
        A = 10.
        n = len(x)
        result = A*n
        for xi in x:
            result += xi**2 - A*np.cos(2*np.pi*xi)
        return result

class sphere(function_ND):
    """
    Multi-dimensional Sphere function
    """

    def __init__(self, Xc=None):
        super().__init__()
        self.Xc = Xc
        if self.Xc is None:
            self.Xc = []

    def call_method(self, *X):
        if len(X)==1 and np.iterable(X[0]):
            X = X[0]
        Nc = len(self.Xc)
        res = 0
        for i, xi in enumerate(X):
            if i < Nc:
                xc = self.Xc[i]
            else:
                xc = 0
            res += (xi - xc) ** 2
        return res**0.5

###########################################################
####################  CostEvaluation  #####################
###########################################################
class CostEvaluation(nrv_function):
    def __init__(self):
        super().__init__()
        self.f_type = "CostEvaluation"

    @staticmethod
    def call_method(self, static_sim) -> float:
        return 1

    def __call__(self, results, **kwargs) -> float:
        return self.call_method(results)


###############################################################
####################### interpolation ######################
###############################################################
class nrv_interp(nrv_function):
    def __init__(
        self,
        X_values,
        Y_values,
        kind="linear",
        dx=0.01,
        interpolator=None,
        dxdy=None,
        scale=None,
        columns=[],
    ):
        """
        Interpolator based on scipy.interpolate
        Parameters
        ----------
        X_values    : (N,) array_like
            A 1-D array of real values
        Y_values    :(…,N,…) array_like
            A N-D array of real values. The length of y along the interpolation
            axis must be equal to the length of x.
        kind        : str
            kind of interpolation which should be used, by default "linear"
            possible values:    - scipy interp1d kinds (see nrv.spy_interp1D_kind)
                                - {'hermite', 'cardinal', 'catmull-rom'} for respective
                                  cubic spline interpollation
        dx
            if minimal distance between 2 consecutive points, by default 0.01
        interpolator
            custumize intepolation function if None kind is use to set the interpolator,
            by default None
        dxdy
            if kind is 'hermite', derivative vector (see scipy.interpolate.CubicHermiteSpline),
            by default None
        scale
            if kind is 'cardinal', scale to use for derivative vector, by default None
        columns       : int or array_like
            columns on which apply the interpolation when class called
            if [] iterpolation done on the whole vector,by default []

        Returns
        -------
        output  :  tuple(3)
            (domain, cell_tag, facet_tag)
        """
        super().__init__()
        self.f_type = "interp"
        # General parameters
        self.X_values = X_values
        self.Y_values = Y_values
        self.N_pts = len(Y_values)
        self.interpolator = interpolator
        self.kind = kind
        self.dx = dx

        # CubicHermiteSpline parameters
        self.dxdy = dxdy

        # Cardinal Spline parameters
        self.scale = scale

        self.columns = columns
        self.__set_interp()

    def __set_interp(self):
        """
        Internal use only, set interpolator
        """
        self.N_pts = len(self.X_values)
        self.__check_X_values()
        self.update_interpolator()

    def __check_X_values(self, dx=None):
        """ """
        S = self.X_values.argsort()
        self.X_values = self.X_values[S]
        self.Y_values = self.Y_values[S]
        if dx is not None:
            self.dx = dx
        for i, x in enumerate(self.X_values):
            if i > 0:
                while x < self.X_values[i - 1]:
                    x += dx
                self.X_values[i] = x

    def update_interpolator(self, kind=None, interpolator=None, dxdy=None, scale=None):
        """ """
        if kind is not None:
            self.kind = kind

        if interpolator is not None:
            self.interpolator = interpolator
        elif self.kind.lower() in spy_interp1D_kind:
            self.interpolator = interp1d(
                self.X_values, self.Y_values, kind=self.kind, fill_value="extrapolate"
            )
        elif self.kind.lower() in ["hermite", "cardinal", "catmull-rom"]:
            if self.scale is None or self.kind == "catmull-rom":
                self.scale = 0.5
            if self.dxdy is None:
                self.dxdy = np.array(
                    [
                        (self.Y_values[1] - self.Y_values[0]) / self.X_values[1]
                        - self.X_values[0]
                    ]
                    + [
                        (self.Y_values[k + 2] - self.Y_values[k])
                        / (self.X_values[k + 2] - self.X_values[k])
                        for k in range(self.N_pts - 2)
                    ]
                    + [
                        (self.Y_values[-1] - self.Y_values[-2])
                        / (self.X_values[-1] - self.X_values[-2])
                    ]
                )
            if self.kind.lower() == "hermite" and dxdy is not None:
                self.dxdy = dxdy
                self.scale = 1
            if self.kind.lower() == "cardinal" and scale is not None:
                self.scale = scale

            self.interpolator = CubicHermiteSpline(
                self.X_values, self.Y_values, self.scale * self.dxdy
            )

    def __call__(self, X):
        if self.columns == []:
            return self.interpolator(X)
        try:
            return self.interpolator(X[self.columns])
        except:
            rise_warning(
                "nrv_interpol: columns out of bound, intepolation done on the whole vector"
            )
            return self.interpolator(X)


class MeshCallBack(nrv_function):
    """ """

    def __init__(self, f=None, axis="x"):
        super().__init__()
        self.f_type = "MeshCallBack"
        self.f = None
        self.axis = axis

        self.set_function(f)

    def set_function(self, f=None):
        if f is None:
            self.f = lambda *x: 1
        elif callable(f):
            self.f = f
        elif isinstance(f, str):
            self.f = lambda x: eval(f)
        else:
            rise_warning(type(f), "Not recognized for MeshCallBack function")

    def __call__(self, dim, tag, x, y, z, lc):
        X = []
        if "x" in self.axis:
            X += [x]
        if "y" in self.axis:
            X += [y]
        if "z" in self.axis:
            X += [z]
        return lc * self.f(*X)



import numpy as np

from ....backend.file_handler import json_dump, json_load, rmv_ext
from ....backend.log_interface import rise_error, rise_warning
from ....backend.NRV_Class import NRV_class


def is_sim_param(X):
    """
    check if an object is a SimParameters, return True if yes, else False

    Parameters
    ----------
    X : object
        object to test

    Returns
    -------
    bool
        True it the type is a SimParameters object
    """
    return isinstance(X, SimParameters)


class SimParameters(NRV_class):
    """
    Class gathering parameters of a FEM FEniCS simulation. It allows to add,
    modify and store domains, boundaries condition and internal boundaries (thin layers)
    Can be saved and used to generate FEMsimulation.
    """

    def __init__(self, D=3, mesh_file="", data=None):
        """
        initialisation of the SimParameters:

        Parameters
        ----------
        D               :int
            dim of the mesh, by default 3
            NB: only 3 is implemented
        mesh_file       :str
            mesh directory and file name: by default ""
        data            :str, dict or SimParameters
            if not None, load SimParameters attribute from data, by default None
            (see SimParameters.load)
        """
        super().__init__()
        self.type = "simparameters"
        self.D = D
        self.mesh_file = rmv_ext(mesh_file)

        # domains
        self.Ndomains = 0
        self.domainsID = []
        self.domains_list = {}

        # boundaries
        self.Nboundaries = 0
        self.boundariesID = []
        self.boundaries_list = {}

        # internal boundaries (thin layer)
        self.inbound = False
        self.Ninboundaries = 0
        self.inboundariesID = []
        self.inboundaries_list = {}

        # gather mat_pty name for domain and internal layer
        self.mat_pty_map = {}

        if data is not None:
            self.load(data)

    def save(self, save=False, fname="SimParameters.json"):
        """
        Return SimParameters as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "SimParameters.json"

        Returns
        -------
        sp_dic : dict
            dictionary containing all information
        """
        sp_dic = {}
        sp_dic["D"] = self.D
        sp_dic["file"] = self.mesh_file
        sp_dic["Ndomains"] = self.Ndomains
        sp_dic["domainsID"] = self.domainsID
        sp_dic["domains"] = self.domains_list
        sp_dic["Nboundaries"] = self.Nboundaries
        sp_dic["boundariesID"] = self.boundariesID
        sp_dic["boundaries"] = self.boundaries_list
        sp_dic["inbound"] = self.inbound
        sp_dic["Ninboundaries"] = self.Ninboundaries
        sp_dic["inboundariesID"] = self.inboundariesID
        sp_dic["inboundaries"] = self.inboundaries_list
        sp_dic["mat_pty_map"] = self.mat_pty_map
        if save:
            json_dump(sp_dic, fname)
        return sp_dic

    def load(self, data):
        """
        Load all SimParameters properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing SimParameters information
        """
        if type(data) == str:
            sp_dic = json_load(data)
        elif is_sim_param(data):
            sp_dic = self.save()
        else:
            sp_dic = data
        self.D = sp_dic["D"]
        self.mesh_file = sp_dic["file"]
        self.Ndomains = sp_dic["Ndomains"]
        self.domainsID = sp_dic["domainsID"]
        self.domains_list = sp_dic["domains"]
        self.Nboundaries = sp_dic["Nboundaries"]
        self.boundariesID = sp_dic["boundariesID"]
        self.boundaries_list = sp_dic["boundaries"]
        self.inbound = sp_dic["inbound"]
        self.Ninboundaries = sp_dic["Ninboundaries"]
        self.inboundariesID = sp_dic["inboundariesID"]
        self.inboundaries_list = sp_dic["inboundaries"]
        self.mat_pty_map = sp_dic["mat_pty_map"]

    def save_SimParameters(self, save=False, fname="SimParameters.json"):
        rise_warning("save_SimParameters is a deprecated method use save")
        return self.save(save=save, fname=fname)

    def load_SimParameters(self, data="SimParameters.json"):
        rise_warning("load_SimParameters is a deprecated method use load")
        self.load(data=data)

    def set_mesh_file(self, new_mesh_file):
        """
        set a new mesh file name

        Parameters
        ----------
        new_mesh_file   :str
            name of the new filename, will be save without the filename extension
        """
        self.mesh_file = new_mesh_file.replace(".gmsh", "").replace(".xdml", "")

    def __update_ID_list(self, lname, ID):
        """
        Internal use only:
        Parameter
        """
        if lname == "domains":
            IDlist = self.domainsID
        elif lname == "inbound":
            IDlist = self.inboundariesID
        elif lname == "bound":
            IDlist = self.boundariesID
        else:
            rise_error("_update_ID_list failed due to unknow list type")

        if IDlist == []:
            if ID is None:
                IDlist = [1]
            else:
                IDlist = [ID]
        else:
            np.sort(IDlist)
            if ID is None:
                IDlist += [IDlist[-1] + 1]
            else:
                if ID in IDlist:
                    return ID
                else:
                    IDlist += [ID]

        if lname == "domains":
            self.domainsID = IDlist
        elif lname == "inbound":
            self.inboundariesID = IDlist
        else:
            self.boundariesID = IDlist
        return IDlist[-1]

    def add_domain(
        self, mesh_domain, mat_pty=None, mat_file=None, mat_perm=None, ID=None
    ):
        """
        add new domain or update if mesh_domain already exists

        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the domain (should be 3D)
        mat_pty         : str, float or list[3]
            Material property, either mat_file or mat_perm
        mat_file        : str
            Material filename (see fmod.material.py) if None use mat_perm,
                by default None
        mat_perm        : float or list[3]
            Material permitivity, if float isotropic mat, if list[3] anisotropic mat
                if None set to 1 (and mat_file is None), by default None
        """
        if mat_pty is not None:
            mat_pty = mat_pty
        elif mat_file is not None:
            mat_pty = mat_file
        elif mat_perm is not None:
            mat_pty = mat_perm
        else:
            mat_pty = 1
            rise_warning(
                "neither mat_file or mat_perm set, set to default permitivity 1 S/m"
            )

        if mesh_domain not in self.domains_list:
            IDdom = self.__update_ID_list("domains", mesh_domain)
            self.domains_list[IDdom] = {
                "mesh_domain": mesh_domain,
                "mat_pty": mat_pty,
                "mixed_domain": [mesh_domain],
            }
        else:
            self.domains_list[mesh_domain]["mat_pty"] = mat_pty
        self.mat_pty_map[mesh_domain] = mat_pty
        self.Ndomains = len(self.domainsID)

    def add_boundary(
        self, mesh_domain, btype, value=None, variable=None, mesh_domain_3D=0, ID=None
    ):
        """
        add new boundary or update if mesh_domain already exists

        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the boundary (should be 2D)
        btype           : str {"Dirichlet", "Neuman"}
            Type of boundary condition either "Dirichlet" to set the
        value           : float
            set the condition to the value
        variable        : str
            name of a python variable from which the boundary value should be taken.
            NB: variable should be place as kwarg in prepare_sim see examples
        mesh_domain_3D  : int
            ID of the mesh of corresponding volume, by default 0
            NB: Most of the time 0 except for LIFE electrode, must be set to the corresponding fascicle)
        ID              : int
            ID of the boundary condition
        """
        IDbound = self.__update_ID_list("bound", mesh_domain)
        if value is not None:
            self.boundaries_list[IDbound] = {
                "mesh_domain": mesh_domain,
                "condition": btype,
                "value": value,
                "mesh_domain_3D": mesh_domain_3D,
            }
        elif variable is not None:
            self.boundaries_list[IDbound] = {
                "mesh_domain": mesh_domain,
                "condition": btype,
                "variable": variable,
                "mesh_domain_3D": mesh_domain_3D,
            }
        else:
            rise_warning("boundary not set, variable or boundary have to be precised")
        self.Nboundaries = len(self.boundariesID)

    def add_inboundary(
        self,
        mesh_domain,
        in_domains,
        thickness,
        mat_pty=None,
        mat_file=None,
        mat_perm=None,
        ID=None,
    ):
        """
        add new internal boundary or update if mesh_domain already exists

        Parameters
        ----------
        mesh_domain     : int
            Mesh ID of the boundary (should be 2D)
        in_domains       : list
            List of mesh IDs of the domain (if 3D group) or internal boundary (if 2D group)
            inside the boundary
        thickness       : float
            thicness of the internal boundary
        mat_pty         : str, float or list[3]
            Material property, either mat_file or mat_perm
        mat_file        : str
            Material filename (see fmod.material.py) if None use mat_perm,
                by default None
        mat_perm        : float or list[3]
            Material permitivity, if float isotropic mat, if list[3] anisotropic mat
                if None set to 1 (and mat_file is None), by default None
        """
        if mat_pty is not None:
            mat_pty = mat_pty
        elif mat_file is not None:
            mat_pty = mat_file
        elif mat_perm is not None:
            mat_pty = mat_perm
        else:
            mat_pty = 1
            rise_warning(
                "neither mat_file or mat_perm set, set to default permitivity 1 S/m"
            )

        self.inbound = True
        if mesh_domain not in self.inboundaries_list:
            IDibound = self.__update_ID_list("inbound", mesh_domain)
            self.inboundaries_list[IDibound] = {
                "mesh_domain": mesh_domain,
                "in_domains": in_domains,
                "thickness": thickness,
                "mat_pty": mat_pty,
            }
            # Compute corresponding domain for mixed space
            for j in self.domains_list:
                domain = self.domains_list[j]
                # Domain inside the boundary
                if domain["mesh_domain"] in in_domains:
                    domain["mixed_domain"] += [domain["mesh_domain"]]
                    if domain["mesh_domain"] == domain["mixed_domain"][0]:
                        domain["mixed_domain"][0] = mesh_domain
                # Domain inside the boundary but surounded by another internal boundary
                elif domain["mesh_domain"] + 1 in in_domains:
                    domain["mixed_domain"] += [domain["mesh_domain"] + 1]
                    if domain["mesh_domain"] == domain["mixed_domain"][0]:
                        domain["mixed_domain"][0] = mesh_domain
                # Domain outside the boundary
                else:
                    domain["mixed_domain"] += [mesh_domain]
        else:
            self.inboundaries_list[mesh_domain]["thickness"] = thickness
            self.inboundaries_list[mesh_domain]["mat_pty"] = mat_pty
        self.mat_pty_map[mesh_domain] = mat_pty
        self.Ninboundaries = len(self.inboundariesID)

    def get_mixedspace_domain(self, i_space=None, i_domain=None):
        """
        TO DO
        """
        if not self.inbound:
            if i_domain is None:
                return self.domains_list
            else:
                return i_domain
        else:
            if i_space is None:
                if i_domain is None:
                    md = [
                        self.domains_list[k]["mixed_domain"] for k in self.domains_list
                    ]
                else:
                    md = self.domains_list[i_domain]["mixed_domain"]
            else:
                if i_domain is None:
                    md = [
                        self.domains_list[k]["mixed_domain"][i_space]
                        for k in self.domains_list
                    ]
                else:
                    md = self.domains_list[i_domain]["mixed_domain"][i_space]
            return md

    def get_mixedspace_mat_pty(self, i_space=None, i_domain=None):
        """
        TO DO
        """
        if self.inbound:
            md = self.get_mixedspace_domain(i_space, i_domain)
            S = np.shape(md)
            if i_space is None:
                if i_domain is None:
                    mmf = [
                        [self.mat_pty_map[md[i][j]] for i in range(S[0])]
                        for j in range(S[1])
                    ]
                else:
                    mmf = [self.mat_pty_map[md[i]] for i in range(S[0])]
            else:
                if i_domain is None:
                    mmf = [self.mat_pty_map[md[i]] for i in range(S[0])]
                else:
                    mmf = self.mat_pty_map[md]
            return mmf

    def get_space_of_domain(self, i_domain):
        """
        TO DO
        """
        if i_domain == 0:
            return 0
        for i in range(len(self.domains_list[i_domain]["mixed_domain"])):
            if self.domains_list[i_domain]["mixed_domain"][i] == i_domain:
                return i
        rise_error("Domain " + str(i_domain) + " is in no space")

    def get_spaces_of_ibound(self, i_ibound):
        """
        retrun the surounding spaces of an internal boundary.
        !! Caution work for NerveMshCreator only : see if it can be generalized
        """
        in_space = self.get_space_of_domain(i_ibound - 1)
        out_space = 0
        if i_ibound > 100:  # Should be a perineurium
            out_space = 0
        else:  # Should be an axon membrane
            for i in self.inboundaries_list:
                if i_ibound in self.inboundaries_list[i]["in_domains"]:
                    out_space = self.get_space_of_domain(i - 1)
        return in_space, out_space

    def print_mixedspace_domain(self):
        doms = self.get_mixedspace_domain()
        print("spaces:   ", [k for k in range(len(doms[0]))])
        for i, dom in enumerate(doms):
            print("domain " + str(i) + ": ", dom)

    def print_mixedspace_mat_pty(self):
        mats = self.get_mixedspace_mat_pty()
        print("domains:   ", [k for k in range(len(mats[0]))])
        for i, mat in enumerate(mats):
            print("space " + str(i) + ": ", mat)



if PostProc_Filtering is not None:
    filter_freq(axon_sim, "V_mem", PostProc_Filtering)
rasterize(axon_sim, "V_mem")
if not (save_V_mem) and self.record_V_mem:
    remove_key(axon_sim, "V_mem", verbose=self.verbose)



import numpy as np
import faulthandler
import traceback

from ..backend.NRV_Class import NRV_class
from ..backend.MCore import MCH, synchronize_processes
from ..backend.log_interface import rise_error, pass_debug_info
from .CostFunctions import CostFunction
from .Optimizers import Optimizer

import sys

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


def cost_function_swarm_from_particle(cost_function_part, **kwargs):
    """
    Generate a cost function for a swarm from a cost function for a particle

    Parameters
    ----------
    cost_function_part    : func
        cost function return a cost (float) from a particle (1-dimensional array)
    verbose      : tupple
        if True, print a progress bar for updated whenby default True

    Returns
    -------
    part        : np.ndarray
        vector of values of the part in the len_part dimension
    """
    def cost_function_swarm(swarm):
        L = len(swarm)
        costs = np.zeros((L))
        for i in range(L):
            particle = swarm[i][:]
            costs[i] = cost_function_part(particle, **kwargs)
            #print("part=", particle, "c=", costs[i] )
        return costs
    return cost_function_swarm

class Problem(NRV_class):
    """
    Problem Class

    A class to describe problems that should be optimized with the NRV Framework.
    The problem should be described with a simulation and a cost, using the object
    CostFunction, and various optimization algorithms can be used to find optimal solution.

    This class is abstract and is not supposed to be used directly by the end user. NRV can
    handle two types of problems:
        - problems where a geometric parameter can be optimized: please refer to ...
        - problems where the waveform can be optimized: please refer to ...
    """

    def __init__(
        self,
        cost_function:CostFunction=None,
        optimizer:Optimizer=None,
        save_problem_results=False,
        problem_fname="optim.json"):
        super().__init__()
        self._CostFunction = cost_function
        self._Optimizer = optimizer
        # For cases where optimisation is done on a swarm(groupe) of particle
        self.swarm_optimizer = False
        self._SwarmCostFunction = None
        self.save_problem_results = save_problem_results
        self.problem_fname = problem_fname
    
    # Handling the CostFunction attribute
    @property
    def costfunction(self):
        """
        Cost function of a Problem,
        the cost function should be a CosFunction object, it should return a scalar.
        NRV function should be prefered"""
        return self._CostFunction

    @costfunction.setter
    def costfunction(self, cf:CostFunction):
        # need to add a verification that the cost function is a scallar and so on
        self._CostFunction = cf

    @costfunction.deleter
    def costfunction(self):
        self._CostFunction = None

    def compute_cost(self, X):
        return self._CostFunction(X)

    # Handling the Optimizer attribute
    @property
    def optimizer(self):
        """
        Optimizer of the problem,
        the Optimizer should be an Optimizer object. It has reference to optimization
        methods and constraints
        """
        return self._Optimizer

    @optimizer.setter
    def optimizer(self, optim:Optimizer):
        self._Optimizer = optim
        self.swarm_optimizer = self._Optimizer.swarm_optimizer

    @optimizer.deleter
    def optmizer(self):
        # self._Optimizer = None
        pass

    # Call method is where the magic happens
    def __call__(self, **kwargs):
        if MCH.do_master_only_work():
            try:
                kwargs = self.__update_saving_parameters(**kwargs)
                if not self.swarm_optimizer:
                    results = self._Optimizer(self._CostFunction, **kwargs)
                else:
                    self._SwarmCostFunction = cost_function_swarm_from_particle(self._CostFunction)
                    results = self._Optimizer(self._SwarmCostFunction, **kwargs)
                MCH.master_broadcasts_to_all({"status":"Completed"})
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except:
                MCH.master_broadcasts_to_all({"status":"Error"})
                rise_error(traceback.format_exc())
            
        elif self.__check_MCore_CostFunction():
            self.__wait_for_simulation()
        else:
            pass
        if MCH.do_master_only_work():
            if self.save_problem_results:
                results.save(save=True, fname=self.problem_fname)
            return results
        else:
            return None

    # Mcore handeling
    def __check_MCore_CostFunction(self):
        return getattr(self._CostFunction, "_MCore_CostFunction", False)

    def __wait_for_simulation(self):
        slave_status = {"status": "Wait"}
        try:
            while slave_status["status"] == "Wait":
                slave_status = MCH.master_broadcasts_to_all(slave_status)
                pass_debug_info(MCH.rank, slave_status)
                sys.stdout.flush()
                if slave_status["status"]=="Simulate":
                    self._CostFunction(slave_status["X"])
                    slave_status["status"] = "Wait"
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        pass_debug_info(MCH.rank, slave_status)

    # additional methods
    def __update_saving_parameters(self, **kwargs):
        """
        internal use only: update the results saving parameters
        and remove the corresponding keys from kwargs
        """
        if "save_problem_results" in kwargs:
            self.save_problem_results = kwargs.pop("save_problem_results")
        if "problem_fname" in kwargs:
            self.problem_fname = kwargs.pop("problem_fname")
        return kwargs


    def context_and_cost(self, context_func, cost_func, residual):
        self.CostFunction = CostFunction(context_func, cost_func, residual)

    def autoset_optimizer(self):
        pass




nrv.rasterize(axon_sim, "V_mem")
nrv.remove_key(axon_sim, "V_mem", verbose=self.verbose)

plt.figure()
plt.scatter(axon_sim["V_mem_raster_time"], axon_sim["V_mem_raster_x_position"])
plt.xlabel("time (ms)")
plt.ylabel("position along the axon($\mu m$)")
title = (
    "Axon "
    + str(k)
    + ", myelination is "
    + str(axon_sim["myelinated"])
    + ", "
    + str(axon_sim["diameter"])
    + " um diameter"
)
plt.title(title)
plt.xlim(0, axon_sim["tstop"])
plt.ylim(0, axon_sim["L"])
if self.save_results:
    plt.tight_layout()
    fig_name = folder_name + "/Activity_axon_" + str(k) + ".pdf"
    plt.savefig(fig_name)
    plt.close()



from setuptools import setup
from pathlib import Path


# CHANGEME VARS
PACKAGE_NAME = 'nrv-py'
DESCRIPTION = 'NeuRon Virtualizer (NRV)'
AUTHOR_NAME = 'Florian Kolbl, Roland Giraud, Louis Regnacq, Thomas Couppey'
PROJECT_URL = "https://github.com/fkolbl/NRV"
this_directory = Path(__file__).parent
long_description = (this_directory / "ReadMe.md").read_text()

'''
# prevent from unnistalled requierements for nrv import
deps = (
    "gmsh",
    "mph",
    "neuron",
    "icecream",
    "numba",
    "mpi4py",
    "scipy",
    "numpy",
    "ezdxf",
    "dolfinx",
    "petsc4py",
    "ufl",
    "gmsh",
    "dolfinx.io",
    "petsc4py.PETSc",
    "dolfinx.fem",
    "dolfinx.fem.petsc",
    "dolfinx.io.utils",
    "scipy.interpolate",
    "scipy.special",
    "dolfinx.io.gmshio",
    "dolfinx.geometry",
    "dolfinx.mesh",
    "numpy.linalg",
    "matplotlib",
    "matplotlib.pyplot",
    "numpy.core",
    "numpy.core._multiarray_umath",
    "matplotlib._path",
    "scipy.stats",
    "scipy.optimize",
    "matplotlib.pylab",
    "pylab",
    "scipy.spatial",
    "scipy.sparse",
    "scipy.sparse.csgraph",
)
for package in deps:
    sys.modules[package] = MagicMock()
'''

setup(
    # meta infos
    name=PACKAGE_NAME,
    author=AUTHOR_NAME,
    description=DESCRIPTION,
    url=PROJECT_URL,
    version="0.9.15",
    long_description=long_description,
    long_description_content_type='text/markdown',

    # architecture
    packages=[
        "nrv",
        "nrv._misc",
        "nrv._misc.OTF_PP",
        "nrv._misc.comsol_templates",
        "nrv._misc.geom",
        "nrv._misc.log",
        "nrv._misc.materials",
        "nrv._misc.mods",
        "nrv._misc.pops",
        "nrv._misc.ppops",
        "nrv._misc.stats",
        "nrv.backend",
        "nrv.fmod",
        "nrv.fmod.FEM",
        "nrv.fmod.FEM.fenics_utils",
        "nrv.fmod.FEM.mesh_creator",
        "nrv.nmod",
        "nrv.optim",
        "nrv.utils",
        "nrv.utils.cell",
        "nrv.utils.fascicle",
    ],
    
    # non python data to keep
    package_data={
        "nrv": ["nrv2calm"],
        "nrv._misc": ["NRV.ini"],
        "nrv._misc.comsol_templates": ["*.mph"],
        "nrv._misc.geom": ["*.dxf", "*.png"],
        "nrv._misc.log": ["NRV.log"],
        "nrv._misc.materials": ["*.mat"],
        "nrv._misc.mods": ["*.mod"],
        "nrv._misc.pops": ["*.pop"],
        "nrv._misc.ppops": ["*.ppop"],
        "nrv._misc.stats": ["*.csv"],
    },
    include_package_data=True,
    # classifiers
    classifiers=[
        'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Natural Language :: English',
    ],
    # dependencies
    install_requires=[
        "gmsh",
        "mph",
        "numpy",
        "scipy",
        "neuron",
        "matplotlib",
        "numba",
        "ezdxf",
        "icecream",
    ],  # external packages as dependencies
    python_requires=">=3.9",
)



################################
## rasterize membrane voltage ##
################################
nrv.rasterize(axon_sim, "V_mem")

###############################
## remove non nevessary data ##
###############################
list_keys = [
    "ID",
    "L",
    "V_mem_raster_position",
    "V_mem_raster_x_position",
    "V_mem_raster_time_index",
    "V_mem_raster_time",
    "myelinated",
    "y",
    "z",
    "diameter",
    "tstop",
    "intra_stim_positions",
    "extracellular_electrode_x",
    "x_nodes",
    "rec",
]

removable_keys = []
for key in axon_sim:
    if not key in list_keys:
        removable_keys += [key]
for key in removable_keys:
    nrv.remove_key(axon_sim, key, verbose=False)

####################################
## check if the axon fired or not ##
####################################
ax_state = axon_state(axon_sim, save=False)
axon_sim["N_AP"] = ax_state["onset number"]
# print(N_AP)

################################
## Save results in a csv file ##
################################
if self.save_results:
    if "in_nerve" in kwargs:  # Nerve Simulation
        file_object = open(self.save_path + "/onset_summary.csv", "a")
        line = str(self.ID)
    else:  # Fascicle Simulation
        file_object = open(folder_name + "/onset_summary.csv", "a")
        line = ""


    line += (
        str(axon_sim["ID"])
        + "\t"
        + str(axon_sim["y"])
        + "\t"
        + str(axon_sim["z"])
        + "\t"
        + str(axon_sim["diameter"])
        + "\t"
        + str(axon_sim["myelinated"])
        + "\t"
        + str(axon_sim["N_AP"])
        + "\n"
    )
    file_object.write(line)
    file_object.close()



"""
Access and modify NRV Parameters
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
from abc import ABCMeta, abstractmethod
from copy import deepcopy

# sys used in an eval
import sys
import numpy as np
from numpy import iterable

from .file_handler import json_dump, json_load
from .log_interface import pass_debug_info

########################################
#           check object               #
########################################


def is_NRV_class(x):
    return isinstance(x, NRV_class)


def is_NRV_class_list(x):
    if iterable(x):
        for xi in x:
            if not is_NRV_class(xi):
                return False
        return True
    return False


def is_NRV_class_dict(x):
    if isinstance(x, dict):
        for xi in x.values():
            if not is_NRV_class(xi):
                return False
        return True
    return False


##########################################
#           check dictionaries           #
##########################################

def is_NRV_object_dict(x):
    return is_NRV_dict(x) or is_NRV_dict_list(x) or is_NRV_dict_dict(x)


def is_NRV_dict(x):
    if isinstance(x, dict):
        if "nrv_type" in x:
            return True
    return False


def is_NRV_dict_list(x):
    if iterable(x):
        if len(x) > 0:
            for xi in x:
                if not (is_NRV_dict(xi)):
                    return False
            return True
    return False


def is_NRV_dict_dict(x):
    if isinstance(x, dict):
        for key in x:
            if not (is_NRV_dict(x[key])):
                return False
        return True
    return False


class NRV_class(metaclass=ABCMeta):
    """
    Instanciate a basic NRV class
    NRV Class are empty shells, defined as abstract classes of which every class in NRV
    should inherite. This enable automatic context backup with save and load methods
    """

    @abstractmethod
    def __init__(self):
        """
        Init method for NRV class
        """
        self.__NRVObject__ = True
        self.nrv_type = self.__class__.__name__
        pass_debug_info(self.nrv_type, " initialized")

    def __del__(self):
        """
        Destructor for NRV class
        """
        pass_debug_info(self.nrv_type, " deleted")


    def save(self, save=False, fname="nrv_save.json", blacklist=[], **kwargs):
        """
        Generic saving method for NRV class instance

        Parameters
        ----------
        save : bool, optional
            If True, save the NRV object in a json file
        fname : str, optional
            Name of the json file
        blacklist : dict, optional
            Dictionary containing the keys to be excluded from the save
        **kwargs : dict, optional
            Additional arguments to be passed to the save method of the NRV object
        """
        key_dic = {}
        for key in self.__dict__:
            if key not in blacklist:
                if is_NRV_class(self.__dict__[key]):
                    key_dic[key] = self.__dict__[key].save(**kwargs)
                elif is_NRV_class_list(self.__dict__[key]):
                    key_dic[key] = []
                    for i in range(len(self.__dict__[key])):
                        key_dic[key] += [self.__dict__[key][i].save(**kwargs)]
                elif is_NRV_class_dict(self.__dict__[key]):
                    key_dic[key] = {}
                    for i in self.__dict__[key]:
                        key_dic[key][i] = self.__dict__[key][i].save(**kwargs)
                else:
                    key_dic[key] = deepcopy(self.__dict__[key])
        if save:
            json_dump(key_dic, fname)
        return key_dic

    def load(self, data, blacklist={}, **kwargs):
        """
        Generic loading method for NRV class instance

        Parameters
        ----------
        data : dict
            Dictionary containing the NRV object
        blacklist : dict, optional
            Dictionary containing the keys to be excluded from the load
        **kwargs : dict, optional
            Additional arguments to be passed to the load method of the NRV object
        """
        if isinstance(data, str):
            key_dic = json_load(data)
        else:
            key_dic = data
        for key in self.__dict__:
            if key in key_dic and key not in blacklist:
                if is_NRV_object_dict(key_dic[key]):
                    self.__dict__[key] = load_any(key_dic[key], **kwargs)
                elif isinstance(self.__dict__[key], np.ndarray):
                    self.__dict__[key] = np.array(key_dic[key])
                elif key_dic[key] == []:
                    self.__dict__[key] = eval(self.__dict__[key].__class__.__name__)()
                else:
                    self.__dict__[key] = key_dic[key]

    def set_parameters(self,**kawrgs):
        for key in kawrgs:
            if key in self.__dict__:
                self.__dict__[key] = kawrgs[key]

    def get_parameters(self):
        return self.__dict__

def load_any(data, **kwargs):
    """loads an object of any kind from a json file

    Args:
        data : _description_

    Returns:
        _type_: _description_
    """
    if isinstance(data, str):
        key_dic = json_load(data)
    else:
        key_dic = data
    # test if NRV class
    if is_NRV_class(key_dic) or is_NRV_class_list(key_dic):
        nrv_obj = key_dic
    # test if NRV dict
    elif is_NRV_dict(key_dic):
        nrv_type = key_dic["nrv_type"]
        nrv_obj = eval('sys.modules["nrv"].' + nrv_type)()
        nrv_obj.load(key_dic, **kwargs)
    elif is_NRV_dict_dict(key_dic):
        nrv_obj = {}
        for key in key_dic:
            nrv_obj[key] = load_any(key_dic[key], **kwargs)
    elif is_NRV_dict_list(key_dic):
        nrv_obj = []
        for i in key_dic:
            nrv_obj += [load_any(i, **kwargs)]
    return nrv_obj



"""
NRV-Multi-Core/Parallel computing handling
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import numpy as np
from .NRV_Singleton import NRV_singleton

try:
    import mpi4py.MPI as mpi

    comm = mpi.COMM_WORLD
    MCore_Flag = True
except ImportError:
    MCore_Flag = False


class Mcore_handler(metaclass=NRV_singleton):
    """
    Class to handle parallel processing (cores, no threads) in NRV2
    """

    def __init__(self, Flag):
        """
        Instantiation of Mcore

        Parameters
        ----------
        Flag    : bool
            should be true if mpi4py is installed, to be handled by user !
        """
        super().__init__()
        self.Flag = Flag
        if self.Flag:
            self.rank = comm.rank
            self.size = comm.size
        else:
            self.rank = 0
            self.size = 1

    def is_alone(self):
        """
        Check if the process is runing alone or if other instances have been launched.

        Returns
        -------
        alone   : bool
            True if the programm is the only instance launched, else False
        """
        return self.size == 1

    def is_master(self):
        """
        Check if the process is master or not whe parallel computing

        Returns
        -------
        master  : bool
            True if multiple instances are launched and the current process is the master (rank 0)
        """
        return not self.is_alone() and self.rank == 0

    def do_master_only_work(self):
        """
        Check if the process is alone or it is the master to perform non splitable job

        Returns
        -------
        only    : bool
            True if the process is alone or if it is the master, else False
        """
        return self.is_alone() or self.is_master()

    def say_hello(self):
        """
        Display a sentence from each process on prompt. For debug only
        """
        if self.is_alone():
            print("Hi, I am the only core launched")
        elif self.is_master():
            print("Hi, I am the master core")
        else:
            print("Hi, I am a slave core, my ID is " + str(self.rank))

    def split_job_from_arrays(self, len_arrays):
        """
        Split an array for parallel independant computing, by sharing independant sub-spaces \
        of array index

        Parameters
        ----------
        len_arrays  : int
            length of the array containing the full job to perform in parallel

        Returns
        -------
        mask    : np.array
            subspace of the array indexes, specific to each instantiation of the programm
        """
        if self.is_alone():
            mask = np.arange(len_arrays)
        else:
            if self.is_master():
                all_indexes = np.arange(len_arrays)
                mask_chunks = np.array_split(all_indexes, self.size, axis=0)
            else:
                mask_chunks = None
            mask = comm.scatter(mask_chunks, root=0)
        return mask

    def split_job_from_arrays_to_slaves(self, len_arrays):
        """
        Split an array for parallel independant computing, by sharing independant sub-spaces \
        of array index, the master gets a table of all jobs to do initialized to False

        Parameters
        ----------
        len_arrays  : int
            length of the array containing the full job to perform in parallel

        Returns
        -------
        mask    : np.array
            subspace of the array indexes, specific to each instantiation of the programm
        """
        if self.is_master():
            all_indexes = np.arange(len_arrays)
            jobs_to_do = np.asarray(np.full(len(all_indexes), False))
            chunks = np.array_split(all_indexes, self.size - 1, axis=0)
            chunks.insert(0, jobs_to_do)
        else:
            chunks = None
        chunk = comm.scatter(chunks, root=0)
        return chunk

    def master_broadcasts_array_to_all(self, var):
        """
        Broadcast an array to all instances of the process (share jobs performed by the master only)

        Parameters
        ----------
        var : np.array
            variable to broadcast, from the master only, esle None

        Returns
        -------
        data    : np.array
            variable broadcasted in all instances
        """
        return self.master_broadcasts_to_all(var)

    def master_broadcasts_to_all(self, var):
        """
        Broadcast an array to all instances of the process (share jobs performed by the master only)

        Parameters
        ----------
        var : np.array or dict
            variable to broadcast, from the master only, esle None

        Returns
        -------
        data    : np.array or dict
            variable broadcasted in all instances
        """
        if self.is_alone():
            data = var
        else:
            if self.rank == 0:
                data = var
            else:
                data = None
            data = comm.bcast(data, root=0)
        return data

    def gather_jobs(self, partial_result):
        """
        Gather the jobs performed by all instances to the master

        Parameters
        ----------
        partial_result  : dict
            individual result from an instance

        Returns
        -------
        result  : dict
            global dict if master or alone, else None
        """
        if self.is_alone():
            final_result = partial_result
        else:
            list_results = comm.gather(partial_result, root=0)
            if self.is_master():
                final_result = list_results[0]
                for i in range(1, len(list_results)):
                    final_result.update(list_results[i])
            else:
                final_result = None
        return final_result

    def gather_jobs_as_array(self, partial_result):
        """
        Gather the jobs performed by all instances to the master

        Parameters
        ----------
        partial_result  : np.array
            individual result from an instance

        Returns
        -------
        result  : np.array
            global array if master or alone, else None
        """
        if self.is_alone():
            final_result = partial_result
        else:
            results = comm.gather(partial_result, root=0)
            if self.is_master():
                final_result = np.concatenate(tuple(results))
            else:
                final_result = None
        return final_result

    def sum_jobs(self, partial_result):
        """
        Gather, sum and broadcast array for conservative results.

        Parameters
        ----------
        partial_result  : np.array
            individual result from an instance

        Returns
        -------
        result  : np.array
            global array as the sum of all partial results from each core,
            returned to all cores
        """
        if self.is_alone():
            final_result = partial_result
        else:
            results = comm.gather(partial_result, root=0)
            # gather and sum results
            sumed_result = None
            if self.is_master():
                list_results = list(results)
                sumed_result = list_results[0]
                for result in list_results[1:]:
                    sumed_result += result
            # broadcast to all
            final_result = self.master_broadcasts_array_to_all(sumed_result)
        return final_result

    def send_data_to_master(self, data):
        """
        Send a dictionary of data directly to the master.

        Parameters
        ----------
        data : dict
            data to send
        """
        comm.send(data, dest=0)

    def recieve_data_from_slave(self):
        """
        Recieve data from anay source

        Returns:
        --------
        data
        """
        data = comm.recv(source=mpi.ANY_SOURCE)
        return data

    def send_back_array_to_dest(self, data, destination):
        """
        Send a numpy array to a slave

        Parameters
        ----------
        data        : np.array
            data to send
        destination : int
            ID of the process to send the data
        """
        comm.send(data, dest=destination)

    def recieve_potential_array_from_master(self):
        """
        Recieve potenatial data from the master as a numpy array

        Parameters
        ----------
        N_points : int
            number of point of the x-coordinate vector
        N_elec   : int
            number of electrode in the simulation
        """
        data = comm.recv(source=0)
        return data

    def send_synchronization_flag(self):
        """
        Blocking collective communication to force all process to synchronize to a specific line of code

        Returns
        -------
        bool
            a flag set to True
        """
        if self.rank == 0:
            Validation_Flag = True
        else:
            Validation_Flag = None
        Validation_Flag = comm.bcast(Validation_Flag, root=0)
        return Validation_Flag


# public interface
MCH = Mcore_handler(MCore_Flag)


def synchronize_processes():
    """
    synchronize all processes, used to wait saving complete before loading

    Returns
    -------
    bool
        a flag set to True
    """
    if not MCH.is_alone():
        comm.Barrier()
    return 0
